import logging
import dask
import numpy as np
import argparse
import warnings
import traceback
import time
import glob
import sys
import os
import subprocess
from astropy.io import fits
from astropy.wcs import FITSFixedWarning
from casatasks import casalog
from dask import delayed
from dask.distributed import Client
from meersolar.meerpipeline.single_image_meerpbcor import get_pbcor_image
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()
warnings.simplefilter("ignore", FITSFixedWarning)


def get_fits_freq(image_file):
    hdr = fits.getheader(image_file)
    keys = hdr.keys()
    if "CTYPE3" in keys and hdr["CTYPE3"] == "FREQ":
        freq = hdr["CRVAL3"]
        return freq
    elif "CTYPE4" in keys and hdr["CTYPE4"] == "FREQ":
        freq = hdr["CRVAL4"]
        return freq
    else:
        print(f"No frequency axis in image: {image_file}.")
        return


def run_pbcor(
    imagename, pbdir, pbcor_dir, apply_parang, jobid=0, ncpu=8, verbose=False
):
    """
    Run single image orimary beam correction

    Parameters
    ----------
    imagename : str
        Imagename
    pbdir : str
        Primary beam directory
    pbcor_dir : str
        Primary beam corrected image directory
    apply_parang : bool
        Apply parallactic angle correction or not
    jobid : int, optional
        Job ID
    ncpu : int, optional
        Number of CPU threads to use
    verbose: bool, optional
        Verbose output

    Returns
    -------
    int
        Success message
    """
    cmd = [
        "run-meer-singlepbcor",
        imagename,
        "--pbdir",
        pbdir,
        "--pbcor_dir",
        pbcor_dir,
        "--ncpu",
        str(ncpu),
        "--jobid",
        str(jobid),
    ]
    if not apply_parang:
        cmd.append("--no_apply_parang")

    if verbose:
        print("Executing:", " ".join(cmd))

    try:
        result = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            check=False,  # Set to True if you want to raise on error
        )
        if verbose or result.returncode != 0:
            print(result.stdout)
        return result.returncode
    except Exception as e:
        print(f"Exception during primary beam correction: {e}")
        return 1


def pbcor_all_images(
    imagedir,
    make_TB=True,
    make_plots=True,
    apply_parang=True,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Correct primary beam of MeerKAT for images in a directory

    Parameters
    ----------
    imagedir : str
        Name of the image directory
    make_TB : bool, optional
        Make brightness temperature map
    make_plots : bool, optional
        Make plots
    apply_parang : bool, optional
        Apply parallactic angle correction
    jobid : int, optional
        Job ID
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    imagedir = imagedir.rstrip("/")
    pbdir = f"{os.path.dirname(imagedir)}/pbdir"
    pbcor_dir = f"{os.path.dirname(imagedir)}/pbcor_images"
    os.makedirs(pbdir, exist_ok=True)
    os.makedirs(pbcor_dir, exist_ok=True)
    successful_pbcor = 0
    try:
        images = glob.glob(f"{imagedir}/*.fits")
        if make_TB:
            tb_dir = f"{os.path.dirname(imagedir)}/tb_images"
            os.makedirs(tb_dir, exist_ok=True)
        if len(images) == 0:
            print(f"No image is present in image directory: {imagedir}")
            return 1
        first_set = []
        remaining_set = []
        freqs = []
        for image in images:
            freq = get_fits_freq(image)
            if freq in freqs:
                remaining_set.append(image)
            else:
                freqs.append(freq)
                first_set.append(image)
        mem_limit = (
            16 * max([os.path.getsize(image) for image in images]) / 1024**3
        )  # In GB
        if len(first_set) > 0:
            if dask_addr is None:
                dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(first_set),
                        dask_dir=pbdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                )
            else:
                _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(first_set),
                        dask_dir=pbdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                        only_cal=True,
                    )
                )
                dask_client=Client(address=dask_addr)
                os.system(f"rm -rf {dask_dir}")
            wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
            tasks = []
            for image in first_set:
                task = delayed(run_pbcor)(
                    image, pbdir, pbcor_dir, apply_parang, jobid=jobid, ncpu=n_threads
                )
                tasks.append(task)
            futures = dask_client.compute(tasks)
            results = list(dask_client.gather(futures))
            dask_client.close()
            if dask_addr is None:
                dask_cluster.close()
                os.system(f"rm -rf {dask_dir}")
            for r in results:
                if r == 0:
                    successful_pbcor += 1
        if len(remaining_set) > 0:
            print(f"Correcting remaining images of different timestamps.")
            if dask_addr is None:
                dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(remaining_set),
                        dask_dir=pbdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                )
            else:
                _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(remaining_set),
                        dask_dir=pbdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                        only_cal=True,
                    )
                )
                dask_client=Client(address=dask_addr)
                os.system(f"rm -rf {dask_dir}")
            wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
            tasks = []
            for image in remaining_set:
                task = delayed(run_pbcor)(
                    image, pbdir, pbcor_dir, apply_parang, jobid=jobid, ncpu=n_threads
                )
                tasks.append(task)
            futures = dask_client.compute(tasks)
            results = list(dask_client.gather(futures))
            dask_client.close()
            if dask_addr is None:
                dask_cluster.close()
                os.system(f"rm -rf {dask_dir}")
            for r in results:
                if r == 0:
                    successful_pbcor += 1

        ############################################
        # Saving fits in helioprojective coordinates
        ############################################
        if successful_pbcor > 0:
            hpcdir = f"{pbcor_dir}/hpcs"
            pbcor_images = glob.glob(f"{pbcor_dir}/*.fits")
            os.makedirs(hpcdir, exist_ok=True)
            print("Saving primary beam corrected images helioprojective coordinates...")
            for image in pbcor_images:
                save_in_hpc(image, outdir=hpcdir)
            if make_plots:
                print("Making plots of primary beam corrected images ...")
                pngdir = f"{pbcor_dir}/pngs"
                os.makedirs(pngdir, exist_ok=True)
                for image in pbcor_images:
                    try:
                        outimages = plot_in_hpc(
                            image,
                            draw_limb=True,
                            extensions=["png"],
                            outdirs=[pngdir],
                        )
                    except BaseException:
                        junkpng = f"{pngdir}/{os.path.basename(image).split('.fits')[0]}.png.junk"
                        os.system(f"touch {junkpng}")

        ####################################
        # Making brightness temperature maps
        ####################################
        if successful_pbcor > 0 and make_TB:
            for pbcor_image in pbcor_images:
                tb_image = (
                    tb_dir
                    + "/"
                    + os.path.basename(pbcor_image).split(".fits")[0]
                    + "_TB.fits"
                )
                generate_tb_map(pbcor_image, outfile=tb_image)

            ############################################
            # Saving fits in helioprojective coordinates
            ###########################################
            hpcdir = f"{tb_dir}/hpcs"
            tb_images = glob.glob(f"{tb_dir}/*.fits")
            os.makedirs(hpcdir, exist_ok=True)
            print("Saving brightness temperature maps helioprojective coordinates...")
            for image in tb_images:
                save_in_hpc(image, outdir=hpcdir)

            if make_plots:
                print("Making plots of brightness temperature maps..")
                pngdir = f"{tb_dir}/pngs"
                os.makedirs(pngdir, exist_ok=True)
                for image in tb_images:
                    outimages = plot_in_hpc(
                        image,
                        draw_limb=True,
                        extensions=["png"],
                        outdirs=[pngdir],
                    )

        #########################################
        # Final calculations
        #########################################
        print(f"Total input images: {len(images)}")
        if successful_pbcor > 0:
            print(f"Total corrected images: {len(pbcor_images)}")
            if make_TB:
                print(f"Total brightness temperatures maps: {len(tb_images)}")
        else:
            print(f"Total corrected images: 0")
        os.system(f"rm -rf {pbdir}/dask-scratch-space")
        return 0
    except Exception as e:
        traceback.print_exc()
        return 1
    finally:
        os.system(f"rm -rf {pbdir}")


def main(
    imagedir,
    workdir="",
    make_TB=True,
    make_plots=True,
    apply_parang=True,
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    start_remote_log=False,
    dask_addr=None,
):
    """
    Primary beam correction of MeerKAT for a sets of images in a directory

    Parameters
    ----------
    imagedir : str
        Image directory
    workdir : str, optional
        Work directory
    make_TB : bool, optional
        Make brightness temperature map or not
    make_plots : bool, optional
        Make png plots
    apply_parang : bool, optional
        Apply parallactic correction or not
    cpu_frac : float,optional
        CPU fraction
    mem_frac : float
        Memory fraction
    logfile : str, optional
        Log file
    jobid : str, optional
        Job ID
    start_remote_log : bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    if workdir == "":
        workdir = imagedir + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "all_pbcor", logfile, jobname=jobname, password=password
            )
    ###########

    try:
        if os.path.exists(imagedir):
            msg = pbcor_all_images(
                imagedir,
                make_TB=make_TB,
                make_plots=make_plots,
                apply_parang=apply_parang,
                jobid=jobid,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
        else:
            print("Please provide correct image directory path.")
            msg = 1
    except Exception:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(imagedir)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Correct all images for MeerKAT full-pol averaged primary beam",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument("imagedir", help="Path to image directory")
    basic_args.add_argument("--workdir", default="", help="Path to work directory")

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--no_make_TB",
        action="store_false",
        dest="make_TB",
        help="Do not generate brightness temperature map",
    )
    adv_args.add_argument(
        "--no_make_plots",
        action="store_false",
        dest="make_plots",
        help="Do not make png plots",
    )
    adv_args.add_argument(
        "--no_apply_parang",
        action="store_false",
        dest="apply_parang",
        help="Do not apply parallactic angle correction",
    )
    adv_args.add_argument(
        "--start_remote_log",
        action="store_false",
        dest="start_remote_log",
        help="Start remote logger",
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac", type=float, default=0.8, help="CPU usage fraction"
    )
    hard_args.add_argument(
        "--mem_frac", type=float, default=0.8, help="Memory usage fraction"
    )
    hard_args.add_argument("--logfile", default=None, help="Path to log file")
    hard_args.add_argument(
        "--jobid", type=int, default=0, help="Job ID for logging and PID tracking"
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        imagedir=args.imagedir,
        workdir=args.workdir,
        make_TB=args.make_TB,
        make_plots=args.make_plots,
        apply_parang=args.apply_parang,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nPrimary beam corrections are done.\n###################\n"
    )
    os._exit(result)



import logging
import tempfile

from regions import ds9 as read_ds9

log = logging.getLogger(__name__)


def load_regions(region_file):
    """
    Parameters
    ----------
    region_file : str
        Region File

    Returns
    -------
    list of CircleSkyRegions
    """

    # kludge because regions cries over "FK5", wants lowercase
    with tempfile.NamedTemporaryFile() as tmpfile, open(region_file) as rf:
        tmpfile.write(rf.read().lower().encode())
        tmpfile.flush()
        include_regions = read_ds9(tmpfile.name)
        log.info(
            "Read %s inclusion region(s) from %s", len(include_regions), region_file
        )

        return include_regions



import os
from dotenv import load_dotenv

homedir = os.environ.get("HOME")
if homedir is None:
    homedir = os.path.expanduser("~")
username = os.getlogin()
cachedir = f"{homedir}/.solarpipe"
ENV_FILE=f"{cachedir}/meersolar_prefect.env"
load_dotenv(dotenv_path=ENV_FILE, override=False)

from casatasks import casalog
import os
try:
    logfile = casalog.logfile()
    os.remove(logfile)
except BaseException:
    pass



import types
import numpy as np
import traceback
import warnings
import copy
import glob
import os
from astropy.io import fits
from astropy.wcs import WCS
from astropy.wcs import FITSFixedWarning
from .basic_utils import *
from .udocker_utils import *

warnings.simplefilter("ignore", category=FITSFixedWarning)


##########################
# Image analysis related
##########################
def create_circular_mask(msname, cellsize, imsize, mask_radius=20):
    """
    Create fits solar mask

    Parameters
    ----------
    msname : str
        Name of the measurement set
    cellsize : float
        Cell size in arcsec
    imsize : int
        Imsize in number of pixels
    mask_radius : float
        Mask radius in arcmin

    Returns
    -------
    str
        Fits mask file name
    """
    try:
        msname = msname.rstrip("/")
        imagename_prefix = (
            os.path.dirname(os.path.abspath(msname))
            + "/"
            + os.path.basename(msname).split(".ms")[0]
            + "_solar"
        )
        wsclean_args = [
            "-quiet",
            "-scale " + str(cellsize) + "asec",
            "-size " + str(imsize) + " " + str(imsize),
            "-nwlayers 1",
            "-niter 0 -name " + imagename_prefix,
            "-channel-range 0 1",
            "-interval 0 1",
        ]
        wsclean_cmd = "wsclean " + " ".join(wsclean_args) + " " + msname
        msg = run_wsclean(wsclean_cmd, "solarwsclean", verbose=False)
        if msg == 0:
            center = (int(imsize / 2), int(imsize / 2))
            radius = mask_radius * 60 / cellsize
            Y, X = np.ogrid[:imsize, :imsize]
            dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)
            mask = dist_from_center <= radius
            os.system(
                "cp -r "
                + imagename_prefix
                + "-image.fits mask-"
                + os.path.basename(imagename_prefix)
                + ".fits"
            )
            os.system("rm -rf " + imagename_prefix + "*")
            data = fits.getdata("mask-" + os.path.basename(imagename_prefix) + ".fits")
            header = fits.getheader(
                "mask-" + os.path.basename(imagename_prefix) + ".fits"
            )
            data[0, 0, ...][mask] = 1.0
            data[0, 0, ...][~mask] = 0.0
            fits.writeto(
                imagename_prefix + "-mask.fits",
                data=data,
                header=header,
                overwrite=True,
            )
            os.system("rm -rf mask-" + os.path.basename(imagename_prefix) + ".fits")
            if os.path.exists(imagename_prefix + "-mask.fits"):
                return imagename_prefix + "-mask.fits"
            else:
                print("Circular mask could not be created.")
                return
        else:
            print("Circular mask could not be created.")
            return
    except Exception as e:
        traceback.print_exc()
        return


def create_circular_mask_array(data, radius):
    """
    Creating circular mask of a Numpy array

    Parameters
    ----------
    data : numpy.array
        2D numpy array
    radius : int
        Radius in pixels

    Returns
    -------
    numpy.array
        Mask array
    """
    shape = data.shape
    center = (shape[0] // 2, shape[1] // 2)
    Y, X = np.ogrid[: shape[0], : shape[1]]
    dist_from_center = (X - center[1]) ** 2 + (Y - center[0]) ** 2
    mask = dist_from_center <= radius**2
    return mask


def calc_solar_image_stat(imagename, disc_size=18):
    """
    Calculate solar image dynamic range

    Parameters
    ----------
    imagename : str
        Fits image name
    disc_size : float, optional
        Solar disc size in arcmin (default : 18)

    Returns
    -------
    float
        Maximum value
    float
        Minimum value
    float
        RMS values
    float
        Total value
    float
        Mean value
    float
        Median value
    float
        RMS dynamic range
    float
        Min-max dynamic range
    """
    data = fits.getdata(imagename)
    header = fits.getheader(imagename)
    pix_size = abs(header["CDELT1"]) * 3600.0  # In arcsec
    radius = int((disc_size * 60) / pix_size)
    if len(data.shape) > 2:
        data = data[0, 0, ...]
    mask = create_circular_mask_array(data, radius)
    masked_data = copy.deepcopy(data)
    masked_data[mask] = np.nan
    unmasked_data = copy.deepcopy(data)
    unmasked_data[~mask] = np.nan
    maxval = float(np.nanmax(unmasked_data))
    minval = float(np.nanmin(masked_data))
    rms = float(np.nanstd(masked_data))
    total_val = float(np.nansum(unmasked_data))
    rms_dyn = float(maxval / rms)
    minmax_dyn = float(maxval / abs(minval))
    mean_val = float(np.nanmean(unmasked_data))
    median_val = float(np.nanmedian(unmasked_data))
    del data, mask, unmasked_data, masked_data
    return (
        round(maxval, 2),
        round(minval, 2),
        round(rms, 2),
        round(total_val, 2),
        round(mean_val, 2),
        round(median_val, 2),
        round(rms_dyn, 2),
        round(minmax_dyn, 2),
    )


def calc_dyn_range(imagename, modelname, residualname, fits_mask=""):
    """
    Calculate dynamic ranges.


    Parameters
    ----------
    imagename : list or str
        Image FITS file(s)
    modelname : list or str
        Model FITS file(s)
    residualname : list ot str
        Residual FITS file(s)
    fits_mask : str, optional
        FITS file mask

    Returns
    -------
    model_flux : float
        Total model flux.
    dyn_range_rms : float
        Max/RMS dynamic range.
    rms : float
        RMS of the image
    """

    def load_data(name):
        return fits.getdata(name)

    def to_list(x):
        return [x] if isinstance(x, str) else x

    imagename = to_list(imagename)
    modelname = to_list(modelname)
    residualname = to_list(residualname)

    use_mask = bool(fits_mask and os.path.exists(fits_mask))
    mask_data = fits.getdata(fits_mask).astype(bool) if use_mask else None

    model_flux, dr1, rmsvalue = 0, 0, 0

    for i in range(len(imagename)):
        img = imagename[i]
        res = residualname[i]
        image = load_data(img)
        residual = load_data(res)
        rms = np.nanstd(residual)
        if use_mask:
            maxval = np.nanmax(image[mask_data])
        else:
            maxval = np.nanmax(image)
        dr1 += maxval / rms if rms else 0
        rmsvalue += rms

    for mod in modelname:
        model = load_data(mod)
        model_flux += np.nansum(model[mask_data] if use_mask else model)

    rmsvalue = rmsvalue / np.sqrt(len(residualname))
    return float(model_flux), round(float(dr1), 2), round(float(rmsvalue), 2)


def generate_tb_map(imagename, outfile=""):
    """
    Function to generate brightness temperature map

    Parameters
    ----------
    imagename : str
        Name of the flux calibrated image
    outfile : str, optional
        Output brightess temperature image name

    Returns
    -------
    str
        Output image name
    """
    if outfile == "":
        outfile = imagename.split(".fits")[0] + "_TB.fits"
    image_header = fits.getheader(imagename)
    image_data = fits.getdata(imagename)
    major = float(image_header["BMAJ"]) * 3600.0  # In arcsec
    minor = float(image_header["BMIN"]) * 3600.0  # In arcsec
    if image_header["CTYPE3"] == "FREQ":
        freq = image_header["CRVAL3"] / 10**9  # In GHz
    elif image_header["CTYPE4"] == "FREQ":
        freq = image_header["CRVAL4"] / 10**9  # In GHz
    else:
        print(f"No frequency information is present in header for {imagename}.")
        return
    TB_conv_factor = (1.222e6) / ((freq**2) * major * minor)
    TB_data = image_data * TB_conv_factor
    image_header["BUNIT"] = "K"
    fits.writeto(outfile, data=TB_data, header=image_header, overwrite=True)
    return outfile


def cutout_image(fits_file, output_file, x_deg=2):
    """
    Cutout central part of the image

    Parameters
    ----------
    fits_file : str
        Input fits file
    output_file : str
        Output fits file name (If same as input, input image will be overwritten)
    x_deg : float, optional
        Size of the output image in degree

    Returns
    -------
    str
        Output image name
    """
    hdu = fits.open(fits_file)[0]
    data = hdu.data  # shape: (nfreq, nstokes, ny, nx)
    header = hdu.header
    wcs = WCS(header)
    _, _, ny, nx = data.shape
    center_x, center_y = nx // 2, ny // 2
    # Get pixel scale (deg/pixel)
    pix_scale_deg = np.abs(header["CDELT1"])
    x_pix = int((x_deg / pix_scale_deg) / 2)
    # Adjust if cutout size exceeds image size
    max_half_x = nx // 2
    max_half_y = ny // 2
    x_pix = min(x_pix, max_half_x)
    y_pix = min(x_pix, max_half_y)  # Assume square pixels
    # Define slice indices
    x0 = center_x - x_pix
    x1 = center_x + x_pix
    y0 = center_y - y_pix
    y1 = center_y + y_pix
    # Slice data
    cutout_data = data[:, :, y0:y1, x0:x1]
    # Update header
    new_header = header.copy()
    new_header["NAXIS1"] = x1 - x0
    new_header["NAXIS2"] = y1 - y0
    new_header["CRPIX1"] -= x0
    new_header["CRPIX2"] -= y0
    # Save
    fits.writeto(output_file, cutout_data, header=new_header, overwrite=True)
    return output_file


def make_timeavg_image(wsclean_images, outfile_name, keep_wsclean_images=True):
    """
    Convert WSClean images into a time averaged image

    Parameters
    ----------
    wsclean_images : list
        List of WSClean images.
    outfile_name : str
        Name of the output file.
    keep_wsclean_images : bool, optional
        Whether to retain the original WSClean images (default: True).

    Returns
    -------
    str
        Output image name.
    """
    timestamps = []
    for i in range(len(wsclean_images)):
        image = wsclean_images[i]
        if i == 0:
            data = fits.getdata(image)
        else:
            data += fits.getdata(image)
        timestamps.append(fits.getheader(image)["DATE-OBS"])
    data /= len(wsclean_images)
    avg_timestamp = average_timestamp(timestamps)
    header = fits.getheader(wsclean_images[0])
    header["DATE-OBS"] = avg_timestamp
    fits.writeto(outfile_name, data=data, header=header, overwrite=True)
    if not keep_wsclean_images:
        for img in wsclean_images:
            os.system(f"rm -rf {img}")
    return outfile_name


def make_freqavg_image(wsclean_images, outfile_name, keep_wsclean_images=True):
    """
    Convert WSClean images into a frequency averaged image

    Parameters
    ----------
    wsclean_images : list
        List of WSClean images.
    outfile_name : str
        Name of the output file.
    keep_wsclean_images : bool, optional
        Whether to retain the original WSClean images (default: True).

    Returns
    -------
    str
        Output image name.
    """
    freqs = []
    for i in range(len(wsclean_images)):
        image = wsclean_images[i]
        if i == 0:
            data = fits.getdata(image)
        else:
            data += fits.getdata(image)
        header = fits.getheader(image)
        if header["CTYPE3"] == "FREQ":
            freqs.append(float(header["CRVAL3"]))
            freqaxis = 3
        elif header["CTYPE4"] == "FREQ":
            freqs.append(float(header["CRVAL4"]))
            freqaxis = 4
    data /= len(wsclean_images)
    if len(freqs) > 0:
        mean_freq = np.nanmean(freqs)
        width = max(freqs) - min(freqs)
        header = fits.getheader(wsclean_images[0])
        if freqaxis == 3:
            header["CRAVL3"] = mean_freq
            header["CDELT3"] = width
        elif freqaxis == 4:
            header["CRAVL4"] = mean_freq
            header["CDELT4"] = width
    fits.writeto(outfile_name, data=data, header=header, overwrite=True)
    if not keep_wsclean_images:
        for img in wsclean_images:
            os.system(f"rm -rf {img}")
    return outfile_name


def make_stokes_wsclean_imagecube(
    wsclean_images, outfile_name, keep_wsclean_images=True
):
    """
    Convert WSClean images into a Stokes cube image.

    Parameters
    ----------
    wsclean_images : list
        List of WSClean images.
    outfile_name : str
        Name of the output file.
    keep_wsclean_images : bool, optional
        Whether to retain the original WSClean images (default: True).

    Returns
    -------
    str
        Output image name.
    """
    stokes = sorted(
        set(
            (
                os.path.basename(i).split(".fits")[0].split(" - ")[-2]
                if " - " in i
                else "I"
            )
            for i in wsclean_images
        )
    )
    valid_stokes = [
        {"I"},
        {"I", "V"},
        {"I", "Q", "U", "V"},
        {"XX", "YY"},
        {"LL", "RR"},
        {"Q", "U"},
        {"I", "Q"},
    ]
    if set(stokes) not in valid_stokes:
        print("Invalid Stokes combination.")
        return
    imagename_prefix = "temp_" + os.path.basename(wsclean_images[0]).split(" - I")[0]
    imagename = imagename_prefix + ".image"
    data, header = fits.getdata(wsclean_images[0]), fits.getheader(wsclean_images[0])
    for img in wsclean_images[1:]:
        data = np.append(data, fits.getdata(img), axis=0)
    header.update(
        {"NAXIS4": len(stokes), "CRVAL4": 1 if "I" in stokes else -5, "CDELT4": 1}
    )
    temp_fits = imagename_prefix + ".fits"
    fits.writeto(outfile_name, data=data, header=header, overwrite=True)
    if not keep_wsclean_images:
        for img in wsclean_images:
            os.system(f"rm -rf {img}")
    return outfile_name


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]



import psutil
import argparse
import traceback
import glob
import sys
import os
from casatasks import casalog
from casatasks import listobs
from meersolar.utils import get_cachedir, drop_cache, SmartDefaultsHelpFormatter

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass


def show_job_status(clean_old_jobs=False):
    """
    Show MeerSOLAR jobs status

    Parameters
    ----------
    clean_old_jobs : bool, optional
        Clean old informations for stopped jobs
    """
    cachedir = get_cachedir()
    try:
        main_pid_files = glob.glob(f"{cachedir}/main_pids_*.txt")
        if len(main_pid_files) == 0:
            print("No MeerSOLAR jobs is running.")
        else:
            print("####################")
            print("MeerSOLAR Job status")
            print("####################")
            for pid_file in main_pid_files:
                with open(pid_file, "r") as f:
                    line = f.read().split(" ")
                jobid = line[0]
                pid = line[1]
                workdir = line[3]
                outdir = line[4]
                if psutil.pid_exists(int(pid)):
                    running = "Running/Waiting"
                else:
                    running = "Done/Stopped"
                print(
                    f"Job ID: {jobid}, Work direcory: {workdir}, Output directory: {outdir}, Status: {running}"
                )
                print(
                    "#########################################################################################"
                )
                if clean_old_jobs and running == "Done/Stopped":
                    os.system(f"rm -rf {pid_file}")
                    if os.path.exists(f"rm -rf {cachedir}/pids/pids_{jobid}.txt"):
                        os.system(f"rm -rf {cachedir}/pids/pids_{jobid}.txt")
    except Exception as e:
        traceback.print_exc()
    finally:
        drop_cache(cachedir)


def cli():
    parser = argparse.ArgumentParser(
        description="Show MeerSOLAR jobs status.",
        formatter_class=SmartDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--show",
        action="store_true",
        dest="show",
        help="Show job status",
    )
    parser.add_argument(
        "--clean_old_jobs",
        action="store_true",
        dest="clean_old_jobs",
        default=False,
        help="Clean old jobs",
    )
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    try:
        args = parser.parse_args()
        if args.show:
            show_job_status(clean_old_jobs=args.clean_old_jobs)
    except Exception as e:
        traceback.print_exc()


if __name__ == "__main__":
    cli()



import psutil
import numpy as np
import traceback
import copy
import glob
import os
from casatools import msmetadata, table
from .basic_utils import *
from .resource_utils import *
from .proc_manage_utils import *
from .ms_metadata import *
from .casatasks import *
from .flagging import *
from .calibration import *
from .imaging import *
from .image_utils import *
from .udocker_utils import *

def intensity_selfcal(
    msname,
    logger,
    selfcaldir,
    cellsize,
    imsize,
    round_number=0,
    uvrange="",
    minuv=0,
    calmode="ap",
    solint="60s",
    refant="1",
    solmode="",
    gaintype="T",
    applymode="calonly",
    threshold=3,
    weight="briggs",
    robust=0.0,
    use_previous_model=False,
    use_solar_mask=True,
    mask_radius=20,
    min_tol_factor=-1,
    ncpu=-1,
    mem=-1,
):
    """
    A single self-calibration round

    Parameters
    ----------
    msname : str
        Name of the measurement set
    logger : logger
        Python logger
    selfcaldir : str
        Self-calibration directory
    cellsize : float
        Cellsize in arcsec
    imsize :  int
        Image pixel size
    round_number : int, optional
        Selfcal iteration number
    uvrange : float, optional
       UV range for calibration
    calmode : str, optional
        Calibration mode ('p' or 'ap')
    solint : str, optional
        Solution intervals
    refant : str, optional
        Reference antenna
    applymode : str, optional
        Solution apply mode (calonly or calflag)
    threshold : float, optional
        Imaging and auto-masking threshold
    weight : str, optional
        Image weighting
    robust : float, optional
        Robust parameter for briggs weighting
    use_previous_model : bool, optional
        Use previous model
    use_solar_mask : bool, optional
        Use solar disk mask or not
    mask_radius : float, optional
        Mask radius in arcminute
    min_tol_factor : float, optional
        Minimum tolerance factor
    ncpu : int, optional
        Number of CPUs to use in WSClean
    mem : float, optional
        Memory usage limit in WSClean

    Returns
    -------
    int
        Success message
    str
        Caltable name
    float
        RMS based dynamic range
    float
        RMS of the image
    str
        Image name
    str
        Model image name
    str
        Residual image name
    """
    limit_threads(n_threads=ncpu)
    from casatasks import gaincal, bandpass, applycal, flagdata, delmod, flagmanager

    try:
        ##################################
        # Setup wsclean params
        ##################################
        if ncpu < 1:
            ncpu = psutil.cpu_count()
        if mem < 0:
            mem = round(psutil.virtual_memory().available / (1024**3), 2)
        msname = msname.rstrip("/")
        if not use_previous_model:
            delmod(vis=msname, otf=True, scr=True)
        prefix = (
            selfcaldir
            + "/"
            + os.path.basename(msname).split(".ms")[0]
            + "_selfcal_present"
        )
        ############################
        # Determining channel blocks
        ############################
        msmd = msmetadata()
        msmd.open(msname)
        times = msmd.timesforspws(0)
        timeres = np.diff(times)
        pos = np.where(timeres > 3 * np.nanmedian(timeres))[0]
        max_intervals = min(1, len(pos))
        freqs = msmd.chanfreqs(0, unit="MHz")
        freqMHz = np.nanmean(freqs)
        freqres = freqs[1] - freqs[0]
        freq_width = calc_bw_smearing_freqwidth(msname)
        nchan = int(freq_width / freqres)
        total_nchan = len(freqs)
        freq = msmd.meanfreq(0, unit="MHz")
        total_time = max(times) - min(times)
        msmd.close()
        chanres = np.diff(freqs)
        chanres /= np.nanmin(chanres)
        pos = np.where(chanres > 1)[0]
        chanrange_list = []
        start_chan = 0
        for i in range(len(pos) + 1):
            if i > len(pos) - 1:
                end_chan = total_nchan
            else:
                end_chan = pos[i] + 1
            if end_chan - start_chan > 10 or len(chanrange_list) == 0:
                chanrange_list.append(f"{start_chan} {end_chan}")
            else:
                last_chanrange = chanrange_list[-1]
                chanrange_list.remove(last_chanrange)
                start_chan = last_chanrange.split(" ")[0]
                chanrange_list.append(f"{start_chan} {end_chan}")
            start_chan = end_chan + 1
        if len(chanrange_list) == 0:
            unflag_chans, flag_chans = get_chans_flag(msname)
            chanrange_list = [f"{min(unflag_chans)} {max(unflag_chans)}"]

        ########################################
        # Scale bias list and channel range list
        ########################################
        scale_bias_list = []
        for chanrange in chanrange_list:
            start_chan = int(chanrange.split(" ")[0])
            end_chan = int(chanrange.split(" ")[-1])
            mid_chan = int((start_chan + end_chan) / 2)
            mid_freq = freqs[mid_chan]
            scale_bias = round(get_multiscale_bias(mid_freq), 2)
            scale_bias_list.append(scale_bias)

        ############################################
        # Merge channel ranges with identical scale bias
        ############################################
        merged_channels = []
        merged_biases = []
        start, end = map(int, chanrange_list[0].split())
        current_bias = scale_bias_list[0]
        for i in range(1, len(chanrange_list)):
            next_start, next_end = map(int, chanrange_list[i].split())
            next_bias = scale_bias_list[i]
            if next_bias == current_bias:
                # Merge ranges (irrespective of contiguity)
                end = next_end
            else:
                # Finalize current group
                merged_channels.append(f"{start} {end}")
                merged_biases.append(current_bias)
                # Start new group
                start, end = next_start, next_end
                current_bias = next_bias
        # Final group
        merged_channels.append(f"{start} {end}")
        merged_biases.append(current_bias)
        chanrange_list = copy.deepcopy(merged_channels)
        scale_bias_list = copy.deepcopy(merged_biases)
        del merged_channels, merged_biases

        ###############################
        # Temporal chunking list
        ###############################
        nintervals = 1
        nchan_list = []
        nintervals_list = []
        for i in range(len(chanrange_list)):
            chanrange = chanrange_list[i]
            ###################################################################
            # Spectral variation is kept fixed at 10% level.
            # Because selfcal is done along temporal axis, where variation matters most
            ###################################################################
            if min_tol_factor <= 0:
                min_tol_factor = 1.0
            nintervals, _ = get_optimal_image_interval(
                msname,
                chan_range=f"{chanrange.replace(' ',',')}",
                temporal_tol_factor=float(min_tol_factor / 100.0),
                spectral_tol_factor=0.1,
                max_nchan=-1,
                max_ntime=max_intervals,
            )
            nchan_list.append(nchan)
            nintervals_list.append(nintervals)

        os.system(f"rm -rf {prefix}*image.fits {prefix}*residual.fits")

        if weight == "briggs":
            weight += " " + str(robust)

        wsclean_args = [
            "-quiet",
            "-scale " + str(cellsize) + "asec",
            "-size " + str(imsize) + " " + str(imsize),
            "-no-dirty",
            "-gridder tuned-wgridder",
            "-weight " + weight,
            "-niter 10000",
            "-mgain 0.85",
            "-nmiter 5",
            "-gain 0.1",
            "-minuv-l " + str(minuv),
            "-j " + str(ncpu),
            "-abs-mem " + str(mem),
            "-no-negative",
            "-auto-mask " + str(threshold + 0.1),
            "-auto-threshold " + str(threshold),
        ]

        ngrid = int(ncpu / 2)
        if ngrid > 1:
            wsclean_args.append("-parallel-gridding " + str(ngrid))

        ################################################
        # Creating and using solar mask
        ################################################
        if use_solar_mask:
            fits_mask = msname.split(".ms")[0] + "_solar-mask.fits"
            if os.path.exists(fits_mask) == False:
                logger.info(f"Creating solar mask of size: {mask_radius} arcmin.\n")
                fits_mask = create_circular_mask(
                    msname, cellsize, imsize, mask_radius=mask_radius
                )
            if fits_mask is not None and os.path.exists(fits_mask):
                wsclean_args.append("-fits-mask " + fits_mask)

        ######################################
        # Running imaging per channel range
        ######################################
        final_image_list = []
        final_model_list = []
        final_residual_list = []

        for i in range(len(chanrange_list)):
            chanrange = chanrange_list[i]
            per_chanrange_wsclean_args = copy.deepcopy(wsclean_args)

            ######################################
            # Multiscale configuration
            ######################################
            start_chan = int(chanrange.split(" ")[0])
            end_chan = int(chanrange.split(" ")[-1])
            chan_number = int((start_chan + end_chan) / 2)
            sun_dia = calc_sun_dia(freqMHz)  # Sun diameter in arcmin
            sun_rad = sun_dia / 2
            multiscale_scales = calc_multiscale_scales(
                msname, 3, chan_number=chan_number, max_scale=sun_rad
            )
            per_chanrange_wsclean_args.append("-multiscale")
            per_chanrange_wsclean_args.append("-multiscale-gain 0.1")
            per_chanrange_wsclean_args.append(
                "-multiscale-scales " + ",".join([str(s) for s in multiscale_scales])
            )
            per_chanrange_wsclean_args.append(
                f"-multiscale-scale-bias {scale_bias_list[i]}"
            )
            if imsize >= 2048 and 4 * max(multiscale_scales) < 1024:
                per_chanrange_wsclean_args.append("-parallel-deconvolution 1024")

            ###################################################################
            # Spectral variation is kept fixed at 10% level.
            # Because selfcal is done along temporal axis, where variation matters most
            ###################################################################
            if len(scale_bias_list) > 1:
                if min_tol_factor <= 0:
                    min_tol_factor = 1.0
                nintervals, _ = get_optimal_image_interval(
                    msname,
                    chan_range=f"{chanrange.replace(' ',',')}",
                    temporal_tol_factor=float(min_tol_factor / 100.0),
                    spectral_tol_factor=0.1,
                )

            #####################################
            # Spectral imaging configuration
            #####################################
            if nchan > 1:
                per_chanrange_wsclean_args.append(f"-channels-out {nchan}")
                per_chanrange_wsclean_args.append("-no-mf-weighting")
                per_chanrange_wsclean_args.append("-join-channels")

            #####################################
            # Temporal imaging configuration
            #####################################
            if nintervals > 1:
                per_chanrange_wsclean_args.append(f"-intervals-out {nintervals}")
            logger.info(f"Spectral chunks: {nchan}, temporal chunks: {nintervals}.")
            temp_prefix = f"{prefix}_chan_{chanrange.replace(' ','_')}"
            per_chanrange_wsclean_args.append(f"-name {temp_prefix}")
            per_chanrange_wsclean_args.append(f"-channel-range {chanrange}")

            if use_previous_model:
                previous_models = glob.glob(f"{temp_prefix}*model.fits")
                if nchan > 1:
                    total_models_expected = (nchan + 1) * nintervals
                else:
                    total_models_expected = (nchan) * nintervals
                if len(previous_models) == total_models_expected:
                    per_chanrange_wsclean_args.append("-continue")
                else:
                    os.system(f"rm -rf {temp_prefix}*")

            wsclean_cmd = (
                "wsclean " + " ".join(per_chanrange_wsclean_args) + " " + msname
            )
            logger.info(f"WSClean command: {wsclean_cmd}\n")
            msg = run_wsclean(wsclean_cmd, "solarwsclean", verbose=False)
            if msg != 0:
                logger.info(f"Imaging is not successful.\n")
            else:
                #####################################
                # Analyzing images
                #####################################
                wsclean_files = {}
                for suffix in ["image", "model", "residual"]:
                    files = glob.glob(temp_prefix + f"*MFS-{suffix}.fits")
                    if not files:
                        files = glob.glob(temp_prefix + f"*{suffix}.fits")
                    wsclean_files[suffix] = files

                wsclean_images = wsclean_files["image"]
                wsclean_models = wsclean_files["model"]
                wsclean_residuals = wsclean_files["residual"]

                final_image = (
                    temp_prefix.replace("present", f"{round_number}") + "_I_image.fits"
                )
                final_model = (
                    temp_prefix.replace("present", f"{round_number}") + "_I_model.fits"
                )
                final_residual = (
                    temp_prefix.replace("present", f"{round_number}")
                    + "_I_residual.fits"
                )

                if len(wsclean_images) == 0:
                    print("No image is made.")
                elif len(wsclean_images) == 1:
                    os.system(f"cp -r {wsclean_images[0]} {final_image}")
                else:
                    final_image = make_timeavg_image(
                        wsclean_images, final_image, keep_wsclean_images=True
                    )
                final_image_list.append(final_image)
                if len(wsclean_models) == 1:
                    os.system(f"cp -r {wsclean_models[0]} {final_model}")
                else:
                    final_model = make_timeavg_image(
                        wsclean_models, final_model, keep_wsclean_images=True
                    )
                final_model_list.append(final_model)
                if len(wsclean_residuals) == 1:
                    os.system(f"cp -r {wsclean_residuals[0]} {final_residual}")
                else:
                    final_residual = make_timeavg_image(
                        wsclean_residuals, final_residual, keep_wsclean_images=True
                    )
                final_residual_list.append(final_residual)

        #########################################
        # Restoring flags if applymode is calflag
        #########################################
        if applymode == "calflag":
            with suppress_casa_output():
                flags = flagmanager(vis=msname, mode="list")
            keys = flags.keys()
            for k in keys:
                if k == "MS":
                    pass
                else:
                    version = flags[0]["name"]
                    if "applycal" in version:
                        try:
                            with suppress_casa_output():
                                flagmanager(
                                    vis=msname, mode="restore", versionname=version
                                )
                                flagmanager(
                                    vis=msname, mode="delete", versionname=version
                                )
                        except BaseException:
                            pass

        #######################################################################
        # Final frequency averaged images for backup or calculating dynamic ranges
        #######################################################################
        final_image = prefix.replace("present", f"{round_number}") + "_I_image.fits"
        final_model = prefix.replace("present", f"{round_number}") + "_I_model.fits"
        final_residual = (
            prefix.replace("present", f"{round_number}") + "_I_residual.fits"
        )
        if len(final_image_list) == 1:
            os.system(f"mv {final_image_list[0]} {final_image}")
        else:
            final_image = make_freqavg_image(
                final_image_list, final_image, keep_wsclean_images=False
            )
        if len(final_model_list) == 1:
            os.system(f"mv {final_model_list[0]} {final_model}")
        else:
            final_model = make_freqavg_image(
                final_model_list, final_model, keep_wsclean_images=False
            )
        if len(final_residual_list) == 1:
            os.system(f"mv {final_residual_list[0]} {final_residual}")
        else:
            final_residual = make_freqavg_image(
                final_residual_list, final_residual, keep_wsclean_images=False
            )
        os.system("rm -rf *psf.fits")
        #####################################
        # Calculating dynamic ranges
        ######################################
        model_flux, rms_DR, rms = calc_dyn_range(
            final_image,
            final_model,
            final_residual,
            fits_mask=fits_mask,
        )
        if model_flux == 0:
            logger.info(f"No model flux.\n")
            return 1, "", 0, 0, "", "", ""

        #####################
        # Perform calibration
        #####################
        bpass_caltable = prefix.replace("present", f"{round_number}") + ".gcal"
        if os.path.exists(bpass_caltable):
            os.system("rm -rf " + bpass_caltable)

        logger.info(
            f"bandpass(vis='{msname}',caltable='{bpass_caltable}',uvrange='{uvrange}',refant='{refant}',solint='{solint},10MHz',minsnr=1,solnorm=True)\n"
        )
        with suppress_casa_output():
            bandpass(
                vis=msname,
                caltable=bpass_caltable,
                uvrange=uvrange,
                refant=refant,
                minsnr=1,
                solint=f"{solint},10MHz",
                solnorm=True,
            )
        if os.path.exists(bpass_caltable) == False:
            logger.info(f"No gain solutions are found.\n")
            return 2, "", 0, 0, "", "", ""

        #########################################
        # Flagging bad gains
        #########################################
        with suppress_casa_output():
            flagdata(
                vis=bpass_caltable, mode="rflag", datacolumn="CPARAM", flagbackup=False
            )
        tb = table()
        tb.open(bpass_caltable, nomodify=False)
        gain = tb.getcol("CPARAM")
        if calmode == "p":
            gain /= np.abs(gain)
        flag = tb.getcol("FLAG")
        gain[flag] = 1.0
        pos = np.where(np.abs(gain) == 0.0)
        gain[pos] = 1.0
        flag *= False
        tb.putcol("CPARAM", gain)
        tb.putcol("FLAG", flag)
        tb.flush()
        tb.close()

        logger.info(
            f"applycal(vis={msname},gaintable=[{bpass_caltable}],interp=['linear,linearflag'],applymode='{applymode}',calwt=[False])\n"
        )
        with suppress_casa_output():
            applycal(
                vis=msname,
                gaintable=[bpass_caltable],
                interp=["linear,linearflag"],
                applymode=applymode,
                calwt=[False],
            )

        #####################################
        # Flag zeros
        #####################################
        with suppress_casa_output():
            flagdata(
                vis=msname,
                mode="clip",
                clipzeros=True,
                datacolumn="corrected",
                flagbackup=False,
            )
        return (
            0,
            bpass_caltable,
            rms_DR,
            rms,
            final_image,
            final_model,
            final_residual,
        )
    except Exception as e:
        print(e)
        traceback.print_exc()
        return 3, "", 0, 0, "", "", ""



import logging
import psutil
import dask
import numpy as np
import argparse
import traceback
import warnings
import time
import sys
import os
from casatasks import casalog
from casatools import msmetadata, ms as casamstool, table
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *
from meersolar.meerpipeline.flagging import single_ms_flag
from meersolar.meerpipeline.import_model import import_fluxcal_models

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def split_casatask(
    msname="", outputvis="", scan="", time_range="", n_threads=-1, dry_run=False
):
    limit_threads(n_threads=n_threads)
    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    with suppress_casa_output():
        from casatasks import split

        split(
            vis=msname,
            outputvis=outputvis,
            scan=scan,
            timerange=time_range,
            datacolumn="data",
            uvrange="0",
            correlation="XX,YY",
        )
    return outputvis


def split_autocorr(
    msname, workdir, scan_list, time_window=-1, cpu_frac=0.8, mem_frac=0.8, dask_addr=None,
):
    """
    Split auto-correlations

    Parameters
    ----------
    msname : str
        Measurement set
    workdir : str
        Working directory
    scan_list : list
        Scan list
    time_window : float, optional
        Time window in seconds from start time of the scan
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    list
        Splited ms list
    """
    msname = msname.rstrip("/")
    scan_size_list=[get_ms_scan_size(msname,scan,only_autocorr=True) for scan in scan_list]
    mem_limit=max(scan_size_list)
    if dask_addr is None:
        dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
            len(scan_list),
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            min_mem_per_job=mem_limit,
        )
    else:
        _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
            len(scan_list),
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            min_mem_per_job=mem_limit,
            only_cal=True,
        )
        dask_client=Client(address=dask_addr)
        os.system(f"rm -rf {dask_dir}")
    wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
    tasks = []
    for scan in scan_list:
        if time_window > 0:
            tb = table()
            tb.open(msname)
            tbsel = tb.query(f"SCAN_NUMBER=={scan}")
            times = tbsel.getcol("TIME")
            tbsel.close()
            tb.close()
            if len(times) == 0:
                continue
            start_time = times[0]
            end_time = start_time + time_window  # add in seconds
            if end_time > max(times):
                end_time = max(times)
            start_time = mjdsec_to_timestamp(start_time, str_format=1)
            end_time = mjdsec_to_timestamp(end_time, str_format=1)
            time_range = f"{start_time}~{end_time}"
        else:
            time_range = ""
        outputvis = workdir + "/autocorr_scan_" + str(scan) + ".ms"
        if os.path.exists(outputvis):
            os.system("rm -rf " + outputvis)
        if os.path.exists(outputvis + ".flagversions"):
            os.system("rm -rf " + outputvis + ".flagversions")
        tasks.append(
            delayed(split_casatask)(
                msname, outputvis, scan, time_range, n_threads=n_threads
            )
        )
    if len(tasks) > 0:
        futures = dask_client.compute(tasks)
        autocorr_mslist = list(dask_client.gather(futures))
    else:
        autocorr_mslist = []
    dask_client.close()
    if dask_addr is None:
        dask_cluster.close()
        os.system(f"rm -rf {dask_dir}")
    return autocorr_mslist


def get_on_off_power(msname="", scale_factor="", ant_list=[], dry_run=False):
    """
    Get noise diode on and off power averaged over antennas

    Parameters
    ----------
    msname : str
        Measurement set name
    scale_factor : numpy.array
        Scaling factor for on-off gain offset in fluxcal scan (shape: npol, nchan)
    ant_list : list
        Antenna id list

    Returns
    -------
    numpy.array
        Spectra of power difference
    """
    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    ######################
    mstool = casamstool()
    mstool.open(msname)
    mstool.select({"antenna1": ant_list, "antenna2": ant_list, "uvdist": [0.0, 0.0]})
    mstool.selectpolarization(["XX", "YY"])
    data_dict = mstool.getdata(["DATA", "FLAG"], ifraxis=True)
    mstool.close()
    del mstool
    data_source = np.abs(data_dict["data"]).astype(np.float32)
    data_source[data_dict["flag"]] = np.nan
    del data_dict
    n_total = data_source.shape[-1]
    n_tstamps = (n_total // 2) * 2
    antslice = slice(min(ant_list), max(ant_list) + 1)
    if data_source[0, 0, 0, 0] > data_source[0, 0, 0, 1]:
        on_idx = slice(0, n_tstamps, 2)
        off_idx = slice(1, n_tstamps, 2)
    else:
        on_idx = slice(1, n_tstamps, 2)
        off_idx = slice(0, n_tstamps, 2)
    ###################################################
    # Averaging along time axis in the antenna chunk
    ###################################################
    diff_source = np.nanmean(
        scale_factor[..., None, None] * data_source[..., on_idx]
        - data_source[..., off_idx],
        axis=-1,
    )
    del data_source
    return diff_source


def get_att_per_ant(cal_msname, source_msname, scale_factor, ant_list=[]):
    """
    Get per antenna attenuatioin array

    Parameters
    ----------
    cal_msname : str
        Fluxcal scan with noise diode
    source_msname : str
        Source scan with noise diode
    scale_factor : numpy.array
        Scaling factor for on-off gain offset in fluxcal scan (shape: npol, nchan)
    ant_list : list, optional
        Antenna list, default: all antennas

    Returns
    -------
    numpy.array
        Attenuation per antenna array. Shape : npol, nchan, nantenna
    """
    if len(ant_list) == 0:
        msmd = msmetadata()
        msmd.open(cal_msname)
        nant = msmd.nantennas()
        msmd.close()
        del msmd
        ant_list = [i for i in range(nant)]
    cal_diff = get_on_off_power(cal_msname, scale_factor, ant_list=ant_list)
    source_diff = get_on_off_power(
        source_msname, scale_factor * 0 + 1, ant_list=ant_list
    )
    att = source_diff / cal_diff
    del source_diff, cal_diff
    return att


def get_power_diff(
    cal_msname="",
    source_msname="",
    on_cal="",
    off_cal="",
    n_threads=-1,
    memory_limit=-1,
    dry_run=False,
):
    """
    Estimate power level difference between alternative correlator dumps.


    Parameters
    ----------
    cal_msname : str
        Fluxcal measurement set
    source_msname : str
        Source measurement set
    on_cal : str
        Noise diode on caltable
    off_cal : str
        Noise diode off caltable
    n_threads : int, optional
        Number of OpenMP threads
    memory_limit : float, optional
        Memory limit in GB

    Returns
    -------
    numpy.array
        Attenuation spectra for both polarizations avergaed over all antennas
    numpy.array
        Attenuation spectra for both polarizations for all antennas
    """
    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    import warnings

    warnings.filterwarnings("ignore")
    starttime = time.time()
    limit_threads(n_threads=n_threads)
    if memory_limit == -1:
        memory_limit = psutil.virtual_memory().available / 1024**3  # GB
    cal_msname = cal_msname.rstrip("/")
    source_msname = source_msname.rstrip("/")
    print(
        f"Estimating power difference from {os.path.basename(cal_msname)} and {os.path.basename(source_msname)}....\n"
    )
    # Get MS metadata
    msmd = msmetadata()
    msmd.open(source_msname)
    nrow = int(msmd.nrows())
    nchan = msmd.nchan(0)
    npol = msmd.ncorrforpol(0)
    nant = msmd.nantennas()
    nbaselines = msmd.nbaselines()
    if nbaselines == 0 or nrow % nbaselines != 0:
        nbaselines += nant
    ntime = int(nrow / nbaselines)
    msmd.close()
    del msmd
    ####################################
    # Calculate mean on-off gain offset
    ####################################
    tb = table()
    tb.open(on_cal)
    ongain = np.abs(tb.getcol("CPARAM")) ** 2
    tb.close()
    tb.open(off_cal)
    offgain = np.abs(tb.getcol("CPARAM")) ** 2
    tb.close()
    del tb
    G = (ongain - offgain) / offgain  # Power offset
    del ongain, offgain
    G_mean = np.nanmean(G, axis=-1)  # Averaged over antennas
    del G
    scale_factor = 1 / (1 + G_mean)
    del G_mean
    ########################################
    # Determining chunk size
    ########################################
    cal_mssize = get_column_size(cal_msname, only_autocorr=True)
    source_mssize = get_column_size(source_msname, only_autocorr=True)
    total_mssize = cal_mssize + source_mssize
    scale_factor_size = nant * ntime * scale_factor.nbytes / (1024.0**3)
    att_ant_array_size = (npol * nchan * nant * 16) / (1024.0**3)
    func_mem = get_on_off_power(dry_run=True)
    per_ant_memory = (scale_factor_size + total_mssize) / nant
    per_ant_memory += att_ant_array_size + func_mem
    nant_per_chunk = min(nant, max(2, int(memory_limit / per_ant_memory)))
    ##############################################
    # Estimating per antenna attenuation in chunks
    ##############################################
    ant_blocks = [
        list(range(i, min(i + nant_per_chunk, nant)))
        for i in range(0, nant, nant_per_chunk)
    ]
    for i, ant_block in enumerate(ant_blocks):
        if i == 0:
            att_ant_array = get_att_per_ant(
                cal_msname, source_msname, scale_factor, ant_list=ant_block
            )
        else:
            att_ant_array = np.append(
                att_ant_array,
                get_att_per_ant(
                    cal_msname, source_msname, scale_factor, ant_list=ant_block
                ),
                axis=-1,
            )
    ######################################
    # Averaging over all antennas
    ######################################
    att = np.nanmedian(att_ant_array, axis=-1)
    return att, att_ant_array


def estimate_att(
    msname,
    workdir,
    noise_on_caltable,
    noise_off_caltable,
    noise_diode_flux_scan,
    valid_target_scans,
    time_window=900,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Estimate attenaution scaling

    Parameters
    ----------
    msname : str
        Measurement set name
    workdir : str
        Working directory
    noise_on_caltable : int
        Caltable with noise diode on
    noise_off_caltable : str
        Caltable with noise diode off
    noise_diode_flux_scan : float
        Fluxcal scan with noise diode
    valid_target_scans : list
        Valid list of target scans
    time_window : float, optional
        Time window in second to use
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    dict
        Dictionary with attenuation spectra for each solar scan
    list
        Attenuation spectra file list
    """
    try:
        print("Estimating attenuation ...")
        ###########################################
        # All auto-corr scans spliting
        ###########################################
        all_scans = [noise_diode_flux_scan] + valid_target_scans
        scan_sizes = []
        for scan in all_scans:
            scan_sizes.append(get_ms_scan_size(msname, int(scan), only_autocorr=True))
        print("Spliting auto-correlation in different scans ...")
        autocorr_mslist = split_autocorr(
            msname,
            workdir,
            all_scans,
            time_window=time_window,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            dask_addr=dask_addr,
        )
        if len(autocorr_mslist) == 0:
            print("No scans splited.")

            return 1, None, None
        ##########################################
        # Flagging on corrected data
        ##########################################
        fluxcal_fields, fluxcal_scans = get_fluxcals(msname)
        badspw = get_bad_chans(msname)
        bad_ants, bad_ants_str = get_bad_ants(msname, fieldnames=fluxcal_fields)
        print("Flagging auto-correlation measurement sets ...")
        mem_limit = max(scan_sizes)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(autocorr_mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(autocorr_mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True
                )
            )
            dask_client=Client(address=dask_addr)
            os.system(f"rm -rf {dask_dir}")
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        tasks = []
        for autocorr_msname in autocorr_mslist:
            tasks.append(
                delayed(single_ms_flag)(
                    autocorr_msname,
                    badspw=badspw,
                    bad_ants_str=bad_ants_str,
                    datacolumn="data",
                    use_tfcrop=True,
                    use_rflag=False,
                    flagdimension="freq",
                    flag_autocorr=False,
                    n_threads=n_threads,
                    memory_limit=mem_limit,
                )
            )
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        for autocorr_msname in autocorr_mslist:
            drop_cache(autocorr_msname)
        att_level = {}
        ########################################
        # Calculating per scan level
        ########################################
        print("Calculating noise-diode power difference ...")
        scan_sizes=[get_column_size(ms) for ms in autocorr_mslist]
        mem_limit=max(scan_sizes)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(valid_target_scans),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(valid_target_scans),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
            )
            dask_client=Client(address=dask_addr)
            os.system(f"rm -rf {dask_dir}")
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        all_scaling_files = []
        filtered_scans = []
        tasks = []
        for scan in valid_target_scans:
            autocorr_msname = f"{workdir}/autocorr_scan_{scan}.ms"
            if autocorr_msname not in autocorr_mslist:
                pass
            tasks.append(
                delayed(get_power_diff)(
                    f"{workdir}/autocorr_scan_{noise_diode_flux_scan}.ms",
                    f"{workdir}/autocorr_scan_{scan}.ms",
                    noise_on_caltable,
                    noise_off_caltable,
                    n_threads=n_threads,
                    memory_limit=mem_limit,
                )
            )
            filtered_scans.append(scan)
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        ##########################################
        # Determining frequencies
        ##########################################
        msmd = msmetadata()
        msmd.open(msname)
        freqs = msmd.chanfreqs(0)
        msmd.close()
        msmd.done()
        del msmd
        ########################
        for i in range(len(filtered_scans)):
            att_value = results[i][0]
            att_ant_array = results[i][1]
            if np.nanmean(att_value) < 0:
                att_value *= -1
                att_ant_array *= -1
            scan = filtered_scans[i]
            att_level[scan] = att_value
            filename = (
                workdir
                + "/"
                + os.path.basename(msname).split(".ms")[0]
                + "_attval_scan_"
                + str(scan)
            )
            att_ant_array_percentage_change = (
                att_value[..., None] - att_ant_array
            ) / att_value[..., None]
            flag_ants = []
            for pol in range(2):
                mean_percentage_change = np.nanmean(
                    att_ant_array_percentage_change[pol, ...], axis=0
                )
                std_percentage_change = np.nanstd(
                    att_ant_array_percentage_change[pol, ...], axis=0
                )
                pos = np.where(
                    np.abs(mean_percentage_change) > 3 * std_percentage_change
                )[0]
                if len(pos) > 0:
                    for i in range(len(pos)):
                        if pos[i] not in flag_ants:
                            flag_ants.append(pos[i])
            np.save(
                filename,
                np.array(
                    [scan, freqs, att_value, flag_ants, att_ant_array], dtype="object"
                ),
            )
            all_scaling_files.append(filename + ".npy")
        return 0, att_level, all_scaling_files
    except Exception as e:
        traceback.print_exc()
        return 1, None, None


def run_noise_cal(
    msname,
    workdir,
    keep_backup=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Perform flux calibration using noise diode

    Parameters
    ----------
    msname : str
        Measurement set
    workdir : str
        Working directory
    keep_backup : bool, optional
        Keep backup
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    dict
        Attenuation values for different scans
    list
        File list saved attenuation values for different scans
    """
    start_time = time.time()
    ncpus = int(psutil.cpu_count() * (1 - cpu_frac))
    limit_threads(n_threads=ncpus)
    from casatasks import split, bandpass

    msname = msname.rstrip("/")
    workdir = workdir.rstrip("/")
    try:
        os.chdir(workdir)
        print("##############################################")
        print("Performing flux calibration using noise-diode.")
        print("##############################################")
        ###################################
        # Determining noise diode cal scans
        ###################################
        fluxcal_fields, fluxcal_scans = get_fluxcals(msname)
        target_scans, cal_scans, f_scans, g_scans, p_scans = get_cal_target_scans(
            msname
        )
        valid_scans = get_valid_scans(msname)
        noise_diode_cal_scan = ""
        for scan in f_scans:
            if scan in valid_scans:
                noise_cal = determine_noise_diode_cal_scan(msname, scan)
                if noise_cal:
                    noise_diode_cal_scan = scan
                    break
        valid_target_scans = []
        for scan in target_scans:
            if scan in valid_scans:
                valid_target_scans.append(scan)
        if noise_diode_cal_scan == "":
            print("No noise diode cal scan is present.")
            print("##################")
            print("Total time taken : ", time.time() - start_time)
            print("##################\n")

            return 1, None, None

        ##############################
        # Split noise cal scan
        ##############################
        print("Spliting auto-correlation in different scans ...")
        noisecal_ms = workdir + "/noisecal.ms"
        if os.path.exists(noisecal_ms):
            os.system("rm -rf " + noisecal_ms)
        if os.path.exists(noisecal_ms + ".flagversions"):
            os.system("rm -rf " + noisecal_ms + ".flagversions")

        with suppress_casa_output():
            split(
                vis=msname,
                outputvis=noisecal_ms,
                scan=noise_diode_cal_scan,
                datacolumn="data",
            )

        ###############################
        # Flagging
        ###############################
        print("Flagging noise cal ms....")
        fluxcal_fields, fluxcal_scans = get_fluxcals(msname)
        badspw = get_bad_chans(msname)
        bad_ants, bad_ants_str = get_bad_ants(msname, fieldnames=fluxcal_fields)
        single_ms_flag(
            noisecal_ms,
            badspw=badspw,
            bad_ants_str=bad_ants_str,
            datacolumn="data",
            use_tfcrop=True,
            use_rflag=False,
            flagdimension="freq",
            flag_autocorr=False,
        )

        ##################################
        # Import models
        ##################################
        print("Importing calibrator models ....")
        fluxcal_result = import_fluxcal_models(
            [noisecal_ms],
            fluxcal_fields,
            fluxcal_scans,
            ncpus=ncpus,
            mem_frac=1 - cpu_frac,
        )

        ##################################
        # Bandpass calibration
        ##################################
        print("Peforming bandpass with noise diode on and off....")
        msmd = msmetadata()
        msmd.open(noisecal_ms)
        times = msmd.timesforscan(noise_diode_cal_scan)
        msmd.close()
        even_times = times[::2]  # Even-indexed timestamps
        odd_times = times[1::2]  # Odd-indexed timestamps
        even_timerange = ",".join(
            [mjdsec_to_timestamp(t, str_format=1) for t in even_times]
        )
        odd_timerange = ",".join(
            [mjdsec_to_timestamp(t, str_format=1) for t in odd_times]
        )
        mstool = casamstool()
        mstool.open(noisecal_ms)
        mstool.select({"antenna1": 1, "antenna2": 1, "time": [times[0], times[1]]})
        dataeven = np.abs(mstool.getdata("DATA", ifraxis=True)["data"])
        mstool.close()
        mstool.open(noisecal_ms)
        mstool.select({"antenna1": 1, "antenna2": 1, "time": [times[1], times[2]]})
        dataodd = np.abs(mstool.getdata("DATA", ifraxis=True)["data"])
        mstool.close()
        if np.nansum(dataeven) > np.nansum(dataodd):
            on_timerange = even_timerange
            off_timerange = odd_timerange
        else:
            on_timerange = odd_timerange
            off_timerange = even_timerange
        oncal = noisecal_ms.split(".ms")[0] + "_on.bcal"
        offcal = noisecal_ms.split(".ms")[0] + "_off.bcal"

        with suppress_casa_output():
            bandpass(
                vis=noisecal_ms,
                caltable=oncal,
                timerange=on_timerange,
                uvrange=">200lambda",
                minsnr=1,
            )
            bandpass(
                vis=noisecal_ms,
                caltable=offcal,
                timerange=off_timerange,
                uvrange=">200lambda",
                minsnr=1,
            )

        #####################################
        # Determine attenuation scaling
        #####################################
        msg, att_level, all_scaling_files = estimate_att(
            msname,
            workdir,
            oncal,
            offcal,
            noise_diode_cal_scan,
            valid_target_scans,
            time_window=900,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            dask_addr=dask_addr,
        )
        if keep_backup:
            print("Backup directory: " + workdir + "/backup")
            os.makedirs(workdir + "/backup", exist_ok=True)
            os.system(f"rm -rf {workdir}/backup/autocorr_scan_*.ms*")
            os.system(
                "mv "
                + noisecal_ms
                + " "
                + oncal
                + " "
                + offcal
                + "  "
                + workdir
                + "/autocorr_scan_*.ms* "
                + workdir
                + "/backup/"
            )
        else:
            os.system(
                "rm -rf "
                + noisecal_ms
                + " "
                + oncal
                + " "
                + offcal
                + " "
                + workdir
                + "/autocorr_scan_*.ms*"
            )
        print("##################")
        print("Total time taken : ", time.time() - start_time)
        print("##################\n")
        return msg, att_level, all_scaling_files
    except Exception as e:
        traceback.print_exc()
        print("##################")
        print("Total time taken : ", time.time() - start_time)
        print("##################\n")
        return 1, None, None


def main(
    msname,
    workdir,
    caldir,
    keep_backup=False,
    start_remote_log=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    dask_addr=None,
):
    """
    Apply calibration solutions to a measurement set.

    Parameters
    ----------
    msname : str
        Path to the target measurement set (MS) to which calibration solutions will be applied.
    workdir : str
        Directory for storing logs, PID files, and intermediate outputs.
    caldir : str
        Directory containing calibration tables (e.g., G, K, B, D Jones terms).
    keep_backup : bool, optional
        If True, keeps a backup of the original MS before applying calibration. Default is False.
    start_remote_log : bool, optional
        Whether to enable remote logging using credentials stored in the `workdir`. Default is False.
    cpu_frac : float, optional
        Fraction of CPU resources to use. Default is 0.8.
    mem_frac : float, optional
        Fraction of system memory to use. Default is 0.8.
    logfile : str or None, optional
        Path to the log file. If None, disables file logging. Default is None.
    jobid : int, optional
        Identifier for tracking the job and saving PID. Default is 0.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    # === Set up workdir ===
    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    if caldir == "" or not os.path.exists(caldir):
        caldir = f"{workdir}/caltables"
    os.makedirs(caldir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "apply_selfcal", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ###########
    try:
        if os.path.exists(msname):
            print("\n###################################")
            print("Starting flux calibration using noise-diode.")
            print("###################################\n")

            msg, att_level, all_scaling_files = run_noise_cal(
                msname,
                workdir,
                keep_backup=keep_backup,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )

            if msg == 0 and all_scaling_files is not None:
                for att_file in all_scaling_files:
                    os.system(f"rm -rf {caldir}/{os.path.basename(att_file)}")
                    os.system("mv " + att_file + " " + caldir)
        else:
            print("Please provide a valid measurement set.")
            msg = 1
    except Exception:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        drop_cache(caldir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Basic calibration using calibrators",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "msname",
        type=str,
        help="Name of measurement set (required positional argument)",
    )
    basic_args.add_argument(
        "--workdir",
        type=str,
        required=True,
        default="",
        help="Working directory (default: auto-created next to MS)",
    )
    basic_args.add_argument(
        "--caldir",
        type=str,
        required=True,
        default="",
        help="Directory for calibration products (default: auto-created in the workdir MS)",
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--keep_backup",
        action="store_true",
        help="Keep backup of measurement set after each round",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac", type=float, default=0.8, help="CPU fraction to use"
    )
    hard_args.add_argument(
        "--mem_frac", type=float, default=0.8, help="Memory fraction to use"
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Path to log file")
    hard_args.add_argument(
        "--jobid", type=str, default="0", help="Job ID for logging and tracking"
    )

    # === Show help if no arguments ===
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        args.msname,
        args.workdir,
        args.caldir,
        keep_backup=args.keep_backup,
        start_remote_log=args.start_remote_log,
        cpu_frac=float(args.cpu_frac),
        mem_frac=float(args.mem_frac),
        logfile=args.logfile,
        jobid=args.jobid,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nNoise diode based flux calibration is finished.\n###################\n"
    )
    os._exit(result)



import types
import psutil
import numpy as np
import traceback
import warnings
import glob
import os
from casatools import msmetadata, ms as casamstool, table
from .basic_utils import *
from .ms_metadata import *
from .imaging import *

#####################################
# Calibration related
#####################################


def merge_caltables(caltables, merged_caltable, append=False, keepcopy=False):
    """
    Merge multiple same type of caltables

    Parameters
    ----------
    caltables : list
        Caltable list
    merged_caltable : str
        Merged caltable name
    append : bool, optional
        Append with exisiting caltable
    keepcopy : bool, opitonal
        Keep input caltables or not

    Returns
    -------
    str
        Merged caltable
    """
    if not isinstance(caltables, list) or len(caltables) == 0:
        print("Please provide a list of caltable.")
        return
    if os.path.exists(merged_caltable) and append:
        pass
    else:
        if os.path.exists(merged_caltable):
            os.system("rm -rf " + merged_caltable)
        if keepcopy:
            os.system("cp -r " + caltables[0] + " " + merged_caltable)
        else:
            os.system("mv " + caltables[0] + " " + merged_caltable)
        caltables.remove(caltables[0])
    if len(caltables) > 0:
        tb = table()
        for caltable in caltables:
            if os.path.exists(caltable):
                tb.open(caltable)
                tb.copyrows(merged_caltable)
                tb.close()
                if not keepcopy:
                    os.system("rm -rf " + caltable)
    return merged_caltable


def get_psf_size(msname, chan_number=-1):
    """
    Function to calculate PSF size in arcsec

    Parameters
    ----------
    msname : str
        Name of the measurement set
    chan_number : int, optional
        Channel number

    Returns
    -------
    float
            PSF size in arcsec
    """
    maxuv_m, maxuv_l = calc_maxuv(msname, chan_number=chan_number)
    psf = np.rad2deg(1.2 / maxuv_l) * 3600.0  # In arcsec
    return round(psf, 2)


def calc_bw_smearing_freqwidth(msname, full_FoV=False, FWHM=True):
    """
    Function to calculate spectral width to procude bandwidth smearing

    Parameters
    ----------
    msname : str
        Name of the measurement set
    full_FoV : bool, optional
        Consider smearing within solar disc or full FoV
    FWHM : bool, optional
        If using full FoV, consider upto FWHM or first null

    Returns
    -------
    float
        Spectral width in MHz
    """
    tb = table()
    tb.open(f"{msname}/SPECTRAL_WINDOW")
    freq = float(tb.getcol("REF_FREQUENCY")[0]) / 10**6
    freqres = float(tb.getcol("CHAN_WIDTH")[0][0]) / 10**6
    tb.close()
    R = 0.9
    if full_FoV:
        fov = calc_field_of_view(msname, FWHM=FWHM)  # In arcsec
    else:
        fov = 2 * calc_sun_dia(np.nanmean(freq)) * 60  # 2 times sun size
    psf = get_psf_size(msname)
    delta_nu = np.sqrt((1 / R**2) - 1) * (psf / fov) * freq
    delta_nu = ceil_to_multiple(delta_nu, freqres)
    return round(delta_nu, 2)


def calc_time_smearing_timewidth(msname, full_FoV=False, FWHM=True):
    """
    Calculate maximum time averaging to avoid time smearing over full FoV.

    Parameters
    ----------
    msname : str
        Measurement set name
    full_FoV : bool, optional
        Consider smearing within solar disc or full FoV
    FWHM : bool, optional
        If using full FoV, consider upto FWHM or first null

    Returns
    -------
    delta_t_max : float
        Maximum allowable time averaging in seconds.
    """
    msmd = msmetadata()
    msmd.open(msname)
    freq_Hz = msmd.chanfreqs(0)[0]
    times = msmd.timesforspws(0)
    msmd.close()
    timeres = times[1] - times[0]
    c = 299792458.0  # speed of light in m/s
    omega_E = 7.2921159e-5  # Earth rotation rate in rad/s
    lam = c / freq_Hz  # wavelength in meters
    freq = freq_Hz / 10**6
    if full_FoV:
        fov = calc_field_of_view(msname, FWHM=FWHM)  # In arcsec
    else:
        fov = 2 * calc_sun_dia(np.nanmean(freq)) * 60  # 2 times sun size
    fov_deg = fov / 3600.0
    fov_rad = np.deg2rad(fov_deg)
    uv, uvlambda = calc_maxuv(msname)
    # Approximate maximum allowable time to avoid >10% amplitude loss
    delta_t_max = lam / (2 * np.pi * uv * omega_E * fov_rad)
    delta_t_max = ceil_to_multiple(delta_t_max, timeres)
    return round(delta_t_max, 2)


def max_time_solar_smearing(msname):
    """
    Max allowable time averaging to avoid solar motion smearing.

    Parameters
    ----------
    msname : str
        Measurement set name

    Returns
    -------
    t_max : float
        Maximum time averaging in seconds.
    """
    omega_sun = 2.5 / (60.0)  # solar apparent motion (2.5 arcsec/min to arcsec/sec)
    psf = get_psf_size(msname)
    t_max = 0.5 * (psf / omega_sun)  # seconds
    return t_max


def delaycal(
    vis="",
    caltable="",
    field="",
    scan="",
    uvrange="",
    refant="",
    refantmode="flex",
    solint="inf",
    combine="scan",
    gaintable=[],
    gainfield=[],
    interp=[],
    dry_run=False,
):
    """
    General delay calibration using CASA, not assuming any point source

    Parameters
    ----------
    vis : str
        Measurement set
    caltable : str
        Caltable name
    field : str, optional
        Field name
    scan : str, optional
        Scan number
    uvrange : str, optional
        UV-range
    refant : str, optional
        Reference antenna (require one reference antenna, can not keep black)
    refantmode : str, optional
        Refant mode
    solint : str, optional
        Solution interval
    combine : str, optional
        Combine data
    gaintable : list, optional
        Previous gaintables
    gainfield : list, optional
        Gain fields
    interp : list, optional
        Interpolation solutions

    Returns
    -------
    str
        Caltable name
    """
    from casatasks import bandpass, gaincal, rerefant

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    try:
        if refant == "":
            print("Provide a reference antenna.")
            return
        warnings.filterwarnings("ignore")
        vis = vis.rstrip("/")
        os.system("rm -rf " + caltable + "*")
        gaincal(
            vis=vis,
            caltable=caltable,
            field=str(field),
            scan=str(scan),
            uvrange="",
            refant=refant,
            solint=solint,
            combine=combine,
            gaintype="K",
            gaintable=gaintable,
            gainfield=gainfield,
            interp=interp,
        )
        bandpass(
            vis=vis,
            caltable=caltable + ".tempbcal",
            field=str(field),
            scan=str(scan),
            uvrange=uvrange,
            refant=refant,
            solint=solint,
            combine=combine,
            solnorm=True,
            gaintable=gaintable,
            gainfield=gainfield,
            interp=interp,
        )
        rerefant(
            vis=vis,
            tablein=caltable + ".tempbcal",
            refant=refant,
            refantmode=refantmode,
        )
        tb = table()
        tb.open(caltable + ".tempbcal/SPECTRAL_WINDOW")
        freq = tb.getcol("CHAN_FREQ").flatten()
        tb.close()
        tb.open(caltable + ".tempbcal")
        gain = tb.getcol("CPARAM")
        flag = tb.getcol("FLAG")
        gain[flag] = np.nan
        tb.close()
        tb.open(caltable, nomodify=False)
        delay_gain = tb.getcol("FPARAM") * 0.0
        delay_flag = tb.getcol("FLAG")
        phase = np.angle(gain)
        for i in range(delay_gain.shape[0]):
            for j in range(delay_gain.shape[2]):
                try:
                    pos = np.where(np.isnan(phase[i, :, j]) == False)[0]
                    if len(pos) > 0:
                        popt, pcov = np.polyfit(
                            2 * np.pi * freq[pos], phase[i, :, j][pos], deg=1
                        )
                        if np.isnan(popt):
                            delay = 0.0
                        else:
                            delay = popt / 10**-9  # Delay in nanosecond
                    else:
                        delay = 0.0
                    delay_gain[i, :, j] = delay
                except Exception:
                    delay_gain[i, :, j] = 0.0
        tb.putcol("FPARAM", delay_gain)
        tb.putcol("FLAG", delay_flag)
        tb.flush()
        tb.close()
        os.system("rm -rf " + caltable + ".tempbcal")
        return caltable
    except Exception as e:
        traceback.print_exc()
        return


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]



import argparse
import sys
from meersolar.utils import *

def cli():
    parser = argparse.ArgumentParser(description="Manage a local Prefect server.")
    subparsers = parser.add_subparsers(dest="command", help="Sub-command help")
    # Start
    start_parser = subparsers.add_parser("start", help="Start the Prefect server")
    start_parser.add_argument(
        "--show-config", action="store_true", help="Display Prefect config after startup"
    )
    # Stop
    subparsers.add_parser("stop", help="Stop the Prefect server")

    # Status
    subparsers.add_parser("status", help="Check if the Prefect server is running")

    # Env
    subparsers.add_parser("save_env", help="Save the Prefect environment to a .env file")

    # Config
    subparsers.add_parser("config", help="Print the current Prefect config")
    
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()

    if args.command == "start":
        start_server(show_config=args.show_config)
    elif args.command == "stop":
        stop_prefect_server()
    elif args.command == "status":
        if prefect_server_status():
            print(f"🟢 Prefect server is running at {SERVER_DASHBOARD}")
        else:
            print("🔴 Prefect server is not running.")
    elif args.command == "save_env":
        save_prefect_env_to_file()
    elif args.command == "config":
        show_prefect_config()
    else:
        parser.print_help()

if __name__=="__main__":
    cli()



import logging
import psutil
import argparse
import requests
import sys
import os
from casatasks import casalog
from datetime import datetime as dt
from parfive import Downloader
from meersolar.utils import (
    get_datadir,
    get_cachedir,
    create_datadir,
    init_udocker,
    initialize_wsclean_container,
    SmartDefaultsHelpFormatter,
    start_server
    
)

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

all_filenames = [
    "3C138_pol_model.txt",
    "udocker-englib-1.2.11.tar.gz",
    "J1939-6342_U_model.txt",
    "UHF_band_cal.npy",
    "L_band_cal.npy",
    "3C286_pol_model.txt",
    "J1939-6342_L_model.txt",
    "J0408-6545_U_model.txt",
    "MeerKAT_antavg_Uband.npz",
    "J0408-6545_L_model.txt",
    "MeerKAT_antavg_Lband.npz",
]


def get_zenodo_file_urls(record_id):
    url = f"https://zenodo.org/api/records/{record_id}"
    response = requests.get(url)
    response.raise_for_status()
    data = response.json()
    return [(f["links"]["self"], f["key"]) for f in data.get("files", [])]


def download_with_parfive(record_id, update=False, output_dir="zenodo_download"):
    print("####################################")
    print("Downloading MeerSOLAR data files ...")
    print("####################################")
    urls = get_zenodo_file_urls(record_id)
    os.makedirs(output_dir, exist_ok=True)
    total_cpu = psutil.cpu_count()
    dl = Downloader(max_conn=min(total_cpu, len(all_filenames)))
    for file_url, filename in urls:
        if filename in all_filenames:
            if os.path.exists(f"{output_dir}/{filename}") == False or update:
                if os.path.exists(f"{output_dir}/{filename}"):
                    os.system(f"rm -rf {output_dir}/{filename}")
                dl.enqueue_file(file_url, path=output_dir, filename=filename)
    results = dl.download()


def init_meersolar_data(update=False, remote_link=None, emails=None):
    """
    Initiate MeerSOLAR data

    Parameters
    ----------
    update : bool, optional
        Update data, if already exists
    remote_link : str, optional
        Remote logger link to save in database
    emails : str, optional
        Email addresses to send remote logger JobID and password
    """
    datadir = get_datadir()
    os.makedirs(datadir, exist_ok=True)
    cachedir = get_cachedir()
    username = os.getlogin()
    linkfile = f"{cachedir}/remotelink_{username}.txt"
    emailfile = f"{cachedir}/emails_{username}.txt"
    if not os.path.exists(linkfile):
        with open(linkfile, "w") as f:
            f.write("")

    if remote_link is not None:
        with open(linkfile, "w") as f:
            f.write(str(remote_link))

    if emails is not None:
        with open(emailfile, "w") as f:
            f.write(str(emails))

    unavailable_files = [
        f for f in all_filenames if not os.path.exists(f"{datadir}/{f}")
    ]

    if unavailable_files or update:
        record_id = "15691548"
        download_with_parfive(record_id, update=update, output_dir=datadir)
        timestr = dt.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        print(f"MeeSOLAR data are updated in: {datadir} at time: {timestr}")


def main(
    init=False,
    prefect_server=True,
    datadir="",
    update=False,
    link=None,
    emails=None,
):
    """
    Initiate MeerSOLAR setup

    Parameters
    ----------
    init : bool, optional
        Initiate setup
    prefect_server : bool, optional
        Initiate prefect server
    datadir : str, optional
        User provided custom data directory
    update : bool, optional
        Update existing data (if corrupted by somehow)
    link : str
        Remote link
    emails : str
        E-mails for notifications
    """
    if init:
        create_datadir(datadir=datadir)
        datadir = get_datadir()
        print(f"MeerSOLAR data directory: {datadir}")
        init_meersolar_data(update=update, remote_link=link, emails=emails)
        print(f"MeerSOLAR data are initiated.")
        init_udocker()
        initialize_wsclean_container()
        print("uDOCKER inititalized")
        if prefect_server:
            start_server()
        return 0
    else:
        return 1


def cli():
    usage = "Initiate MeerSOLAR data"
    parser = argparse.ArgumentParser(
        description=usage, formatter_class=SmartDefaultsHelpFormatter
    )
    parser.add_argument("--init", action="store_true", help="Initiate data")
    parser.add_argument(
        "--datadir", type=str, default="", help="User provided data directory"
    )
    parser.add_argument(
        "--no_prefect_server", action="store_false", dest="init_prefect_server", help="Do not inititate prefect server"
    )
    parser.add_argument("--update", action="store_true", help="Update existing data")
    parser.add_argument(
        "--remotelink", dest="link", default=None, help="Set remote log link"
    )
    parser.add_argument(
        "--emails",
        dest="emails",
        default=None,
        help="Email addresses (comma seperated) to send Job ID and password for remote logger",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()

    msg = main(
        init=args.init,
        datadir=args.datadir,
        prefect_server=args.init_prefect_server,
        update=args.update,
        link=args.link,
        emails=args.emails,
    )
    return msg


if __name__ == "__main__":
    msg = cli()
    os._exit(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from contextlib import ExitStack
import warnings

from dask.array import PerformanceWarning
from loguru import logger as log
import sys


try:
    import dask
    import dask.array as da
    from daskms import xds_from_ms, xds_from_table, xds_to_table
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None

from africanus.coordinates.dask import radec_to_lm
from africanus.rime.dask import wsclean_predict
from africanus.util.dask_util import EstimatingProgressBar
from africanus.util.requirements import requires_optional

import meersolar.crystalball.logger_init  # noqa
from meersolar.crystalball.budget import get_budget
from meersolar.crystalball.filtering import select_field_id, filter_datasets
from meersolar.crystalball.ms import ms_preprocess
from meersolar.crystalball.region import load_regions
from meersolar.crystalball.wsclean import import_from_wsclean, WSCleanModel


def create_parser():
    p = argparse.ArgumentParser()
    p.add_argument("ms", help="Input .MS file.")
    p.add_argument(
        "-sm",
        "--sky-model",
        default="sky-model.txt",
        help="Name of file containing the sky model. " "Default is 'sky-model.txt'",
    )
    p.add_argument(
        "-o",
        "--output-column",
        default="MODEL_DATA",
        help="Output visibility column. Default is '%(default)s'",
    )
    p.add_argument(
        "-f",
        "--field",
        type=str,
        help="The field name or id to be predicted. "
        "If not provided, only a single field "
        "may be present in the MS",
    )
    p.add_argument(
        "-rc",
        "--row-chunks",
        type=int,
        default=0,
        help="Number of rows of input MS that are processed in "
        "a single chunk. If 0 it will be set automatically. "
        "Default is 0.",
    )
    p.add_argument(
        "-mc",
        "--model-chunks",
        type=int,
        default=0,
        help="Number of sky model components that are processed in "
        "a single chunk. If 0 it wil be set automatically. "
        "Default is 0.",
    )
    p.add_argument(
        "-w",
        "--within",
        type=str,
        help="Optional. Give JS9 region file. Only sources within "
        "those regions will be included.",
    )
    p.add_argument(
        "-po",
        "--points-only",
        action="store_true",
        help="Select only point-type sources.",
    )
    p.add_argument(
        "-ns",
        "--num-sources",
        type=int,
        default=0,
        metavar="N",
        help="Select only N brightest sources.",
    )
    p.add_argument(
        "-j",
        "--num-workers",
        type=int,
        default=0,
        metavar="N",
        help="Explicitly set the number of worker threads.",
    )
    p.add_argument(
        "-mf",
        "--memory-fraction",
        type=float,
        default=0.1,
        help="Fraction of system RAM that can be used. "
        "Used when setting automatically the "
        "chunk size. Default in 0.1.",
    )

    return p


def support_tables(args, tables):
    """
    Parameters
    ----------
    args : object
        Script argument objects
    tables : list of str
        List of support tables to open

    Returns
    -------
    table_map : dict of :class:`xarray.Dataset`
        {name: dataset}
    """
    return {
        t: [
            ds.compute()
            for ds in xds_from_table("::".join((args.ms, t)), group_cols="__row__")
        ]
        for t in tables
    }


def fill_correlations(vis, pol):
    """
    Expands single correlation produced by wsclean_predict to the
    full set of correlations.

    Parameters
    ----------
    vis : :class:`dask.array.Array`
        dask array of visibilities of shape :code:`(row, chan, 1)`
    pol : :class:`xarray.Dataset`
        MS Polarisation dataset.

    Returns
    -------
    vis : :class:`dask.array.Array`
        dask array of visibilities of shape :code:`(row, chan, corr)`
    """

    corrs = pol.NUM_CORR.data[0]

    assert vis.ndim == 3

    if corrs == 1:
        return vis
    elif corrs == 2:
        vis = da.concatenate([vis, vis], axis=2)
        return vis.rechunk({2: corrs})
    elif corrs == 4:
        zeros = da.zeros_like(vis)
        vis = da.concatenate([vis, zeros, zeros, vis], axis=2)
        return vis.rechunk({2: corrs})
    else:
        raise ValueError("MS Correlations %d not in (1, 2, 4)" % corrs)


def source_model_to_dask(source_model, chunks):
    # Create chunked dask arrays from wsclean model arrays
    sm = source_model

    radec_chunks = (chunks,) + sm.radec.shape[1:]
    spi_chunks = (chunks,) + sm.spi.shape[1:]
    gauss_chunks = (chunks,) + sm.gauss_shape.shape[1:]

    return WSCleanModel(
        da.from_array(sm.source_type, chunks=chunks),
        da.from_array(sm.radec, chunks=radec_chunks),
        da.from_array(sm.flux, chunks=chunks),
        da.from_array(sm.spi, chunks=spi_chunks),
        da.from_array(sm.ref_freq, chunks=chunks),
        da.from_array(sm.log_poly, chunks=chunks),
        da.from_array(sm.gauss_shape, chunks=gauss_chunks),
    )


def predict():
    # Parse application args
    args = create_parser().parse_args([a for a in sys.argv[1:]])

    with ExitStack() as stack:
        # Set up dask ThreadPool prior to any application dask calls
        if args.num_workers:
            stack.enter_context(dask.config.set(num_workers=args.num_workers))

        # Run application script
        return _predict(args)


@requires_optional("dask.array", "daskms", opt_import_error)
def _predict(args):
    # get inclusion regions
    include_regions = load_regions(args.within) if args.within else []

    # Import source data from WSClean component list
    # See https://sourceforge.net/p/wsclean/wiki/ComponentList
    source_model = import_from_wsclean(
        args.sky_model,
        include_regions=include_regions,
        point_only=args.points_only,
        num=args.num_sources or None,
    )

    # Add output column if it isn't present
    ms_rows, ms_datatype = ms_preprocess(args)

    # Get the support tables
    tables = support_tables(
        args, ["FIELD", "DATA_DESCRIPTION", "SPECTRAL_WINDOW", "POLARIZATION"]
    )

    field_ds = tables["FIELD"]
    ddid_ds = tables["DATA_DESCRIPTION"]
    spw_ds = tables["SPECTRAL_WINDOW"]
    pol_ds = tables["POLARIZATION"]

    max_num_chan = max([ss.NUM_CHAN.data[0] for ss in spw_ds])
    max_num_corr = max([ss.NUM_CORR.data[0] for ss in pol_ds])

    # Perform resource budgeting
    nsources = source_model.source_type.shape[0]
    args.row_chunks, args.model_chunks = get_budget(
        nsources, ms_rows, max_num_chan, max_num_corr, ms_datatype, args
    )

    source_model = source_model_to_dask(source_model, args.model_chunks)

    # List of write operations
    writes = []

    datasets = xds_from_ms(
        args.ms,
        columns=["UVW", "ANTENNA1", "ANTENNA2", "TIME"],
        group_cols=["FIELD_ID", "DATA_DESC_ID"],
        chunks={"row": args.row_chunks},
    )

    field_id = select_field_id(field_ds, args.field)

    for xds in filter_datasets(datasets, field_id):
        # Extract frequencies from the spectral window associated
        # with this data descriptor id
        field = field_ds[xds.attrs["FIELD_ID"]]
        ddid = ddid_ds[xds.attrs["DATA_DESC_ID"]]
        spw = spw_ds[ddid.SPECTRAL_WINDOW_ID.data[0]]
        pol = pol_ds[ddid.POLARIZATION_ID.data[0]]
        frequency = spw.CHAN_FREQ.data[0]

        lm = radec_to_lm(source_model.radec, field.PHASE_DIR.data[0][0])

        with warnings.catch_warnings():
            # Ignore dask chunk warnings emitted when going from 1D
            # inputs to a 2D space of chunks
            warnings.simplefilter("ignore", category=PerformanceWarning)
            vis = wsclean_predict(
                xds.UVW.data,
                lm,
                source_model.source_type,
                source_model.flux,
                source_model.spi,
                source_model.log_poly,
                source_model.ref_freq,
                source_model.gauss_shape,
                frequency,
            )

        vis = fill_correlations(vis, pol)

        log.info(
            "Field {0} DDID {1:d} rows {2} chans {3} corrs {4}",
            field.NAME.values[0],
            xds.DATA_DESC_ID,
            vis.shape[0],
            vis.shape[1],
            vis.shape[2],
        )

        # Assign visibilities to MODEL_DATA array on the dataset
        xds = xds.assign(**{args.output_column: (("row", "chan", "corr"), vis)})
        # Create a write to the table
        write = xds_to_table(xds, args.ms, [args.output_column])
        # Add to the list of writes
        writes.append(write)

    with ExitStack() as stack:
        if sys.stdout.isatty():
            # Default progress bar in user terminal
            stack.enter_context(EstimatingProgressBar())
        else:
            # Log progress every 5 minutes
            stack.enter_context(EstimatingProgressBar(minimum=2 * 60, dt=5))

        # Submit all graph computations in parallel
        dask.compute(writes)

    log.info("Finished")



from loguru import logger as log
import psutil
import numpy as np


def get_budget(
    nr_sources,
    nr_rows,
    nr_chans,
    nr_corrs,
    data_type,
    cb_args,
    fudge_factor=1.25,
    row2source_ratio=100,
):
    systmem = float(psutil.virtual_memory()[0])
    if not cb_args.num_workers:
        nrthreads = psutil.cpu_count()
    else:
        nrthreads = cb_args.num_workers

    log.info("-" * 50)
    log.info("Budgeting")
    log.info("-" * 50)
    log.info("system RAM = {0:.2f} GB", systmem / 1024**3)
    log.info("nr of logical CPUs = {0:d}", nrthreads)
    log.info("nr sources = {0:d}", nr_sources)
    log.info("nr rows    = {0:d}", nr_rows)
    log.info("nr chans   = {0:d}", nr_chans)
    log.info("nr corrs   = {0:d}", nr_corrs)

    data_type = {"complex": "complex64", "dcomplex": "complex128"}[data_type]
    data_bytes = np.dtype(data_type).itemsize
    bytes_per_row = nr_chans * nr_corrs * data_bytes
    memory_per_row = bytes_per_row * fudge_factor

    if cb_args.model_chunks and cb_args.row_chunks:
        rows_per_chunk = cb_args.row_chunks
        sources_per_chunk = cb_args.model_chunks
        strat_type = "(user settings)"
    elif not cb_args.model_chunks and not cb_args.row_chunks:

        if cb_args.memory_fraction > 1 or cb_args.memory_fraction <= 0:
            raise ValueError(
                "The memory fraction must be a number in the "
                "interval (0,1]. You have set it to {0:f} "
                ".".format(cb_args.memory_fraction)
            )

        allowed_rows_per_thread = (
            systmem * cb_args.memory_fraction / (memory_per_row * nrthreads)
        )
        rows_per_chunk = int(min(nr_rows, allowed_rows_per_thread))
        sources_per_chunk = int(min(nr_sources, rows_per_chunk / row2source_ratio))
        strat_type = "(auto settings)"
    else:
        raise ValueError(
            "For now you must set both row and source chunk, or "
            "leave both unset (=0); "
            "you cannot set only one of them."
        )

    log.info("sources per chunk = {0:.0f} {1}", sources_per_chunk, strat_type)
    log.info("rows per chunk    = {0:.0f} {1}", rows_per_chunk, strat_type)

    memory_usage = rows_per_chunk * memory_per_row * nrthreads + sources_per_chunk
    log.info("expected memory usage = {0:.2f} GB", memory_usage / 1024**3)

    return rows_per_chunk, sources_per_chunk



import os
import types
import subprocess
import time
import socket
import signal
import argparse
import toml
from pathlib import Path
from meersolar.utils.basic_utils import *

# === CONFIG ===
cachedir = get_cachedir()
PREFECT_HOME = f"{cachedir}/prefect_home"
os.makedirs(PREFECT_HOME,exist_ok=True)
DB_URL = f"sqlite+aiosqlite:///{PREFECT_HOME}/prefect.db"
LOG_FILE = os.path.join(PREFECT_HOME, "server.log")
profile_path=os.path.join(PREFECT_HOME, "profiles.toml")
memo_path=os.path.join(PREFECT_HOME, "memo_store.toml")
storage=os.path.join(PREFECT_HOME, "storage")
os.makedirs(storage,exist_ok=True)
ENV_FILE = os.path.join(cachedir, "meersolar_prefect.env")
SERVER_HOST = "0.0.0.0"
SERVER_PORT = "4200"
SERVER_URL = f"http://{SERVER_HOST}:{SERVER_PORT}/api"
SERVER_DASHBOARD = f"http://{SERVER_HOST}:{SERVER_PORT}/dashboard"
profile_name="solarpipe"
pid_file = os.path.join(PREFECT_HOME, "server.pid")
logging_path=os.path.join(PREFECT_HOME,"logging.yml")


def write_prefect_profile():
    """
    Save prefect profile
    """
    # Load existing TOML config or start new
    if os.path.exists(profile_path):
        data = toml.load(profile_path)
    else:
        data = {}
    # Set active profile
    data["active"] = profile_name
    # Set config under [profiles.<profile_name>]
    if "profiles" not in data:
        data["profiles"] = {}
    data["profiles"][profile_name] = {
        "PREFECT_API_URL": SERVER_URL,
        "PREFECT_HOME": PREFECT_HOME,
        "PREFECT_API_DATABASE_CONNECTION_URL": DB_URL,
    }
    with open(profile_path, "w") as f:
        toml.dump(data, f)
    print(f"✅ Prefect profile '{profile_name}' written to {profile_path}")
    

def prefect_server_status():
    """
    Get prefect server status
    """
    try:
        with socket.create_connection((SERVER_HOST, SERVER_PORT), timeout=2):
            return True
    except OSError:
        return False


def get_prefect_env():
    """
    Get environment variables of prefect 
    """
    env = os.environ.copy()
    env["PREFECT_HOME"] = PREFECT_HOME
    env["PREFECT_API_MODE"] = "server"
    env["PREFECT_API_DATABASE_CONNECTION_URL"] = DB_URL
    env["PREFECT_SERVER_ALLOW_EPHEMERAL_MODE"] = "false"
    env["PREFECT_API_URL"]=SERVER_URL
    env["PREFECT_PROFILE"] = profile_name
    env["PREFECT_PROFILES_PATH"]=profile_path
    env["PREFECT_LOCAL_STORAGE_PATH"]=storage
    env["PREFECT_LOGGING_SETTINGS_PATH"]=logging_path
    env["PREFECT_MEMO_STORE_PATH"]=memo_path
    return env


def save_prefect_env_to_file():
    """
    Save current Prefect server env config to a .env file for reuse.
    """
    with open(ENV_FILE, "w") as f:
        f.write(f"PREFECT_HOME={PREFECT_HOME}\n")
        f.write("PREFECT_API_MODE=server\n")
        f.write(f"PREFECT_API_DATABASE_CONNECTION_URL={DB_URL}\n")
        f.write("PREFECT_SERVER_ALLOW_EPHEMERAL_MODE=false\n")
        f.write(f"PREFECT_API_URL={SERVER_URL}\n")
        f.write(f"PREFECT_PROFILE={profile_name}\n")
        f.write(f"PREFECT_PROFILES_PATH={profile_path}\n")
        f.write(f"PREFECT_LOCAL_STORAGE_PATH={storage}\n")
        f.write(f"PREFECT_LOGGING_SETTINGS_PATH={logging_path}\n")
        f.write(f"PREFECT_MEMO_STORE_PATH={memo_path}\n")
    print(f"📄 Saved Prefect server environment to {ENV_FILE}")
    if os.path.exists(f"{cachedir}/prefect.dashboard") is not True:
        with open(f"{cachedir}/prefect.dashboard","w") as f:
            f.write(f"{SERVER_DASHBOARD}")
    write_prefect_profile()


def start_server(show_config=False):
    """
    Start prefect server if it is not running
    """
    if prefect_server_status():
        print(f"🟢 Prefect server is already running at {SERVER_DASHBOARD}")
        if os.path.exists(f"{cachedir}/prefect.dashboard") is not True:
            with open(f"{cachedir}/prefect.dashboard","w") as f:
                f.write(f"{SERVER_DASHBOARD}")
        if show_config:
            show_prefect_config()
        os.makedirs(PREFECT_HOME, exist_ok=True)
        save_prefect_env_to_file()
        return 0
    print("🚀 Starting Prefect server...")
    if os.path.exists(pid_file):
        stop_prefect_server()
    with open(LOG_FILE, "w") as f:
        server_proc = subprocess.Popen(
            ["prefect", "server", "start", "--host",SERVER_HOST,"--port",SERVER_PORT],
            stdout=f,
            stderr=subprocess.STDOUT,
            env=get_prefect_env(),
        )
    server_started=False
    for _ in range(30):  # wait up to 30s for the server to respond
        if prefect_server_status():
            if show_config:
                show_prefect_config()
            server_started=True
            break
        time.sleep(1)
    if server_started:
        with open(pid_file, "w") as pf:
            pf.write(str(server_proc.pid))
        os.makedirs(PREFECT_HOME, exist_ok=True)
        save_prefect_env_to_file()
        print(f"✅ Prefect server is now running at {SERVER_DASHBOARD}")
        if os.path.exists(f"{cachedir}/prefect.dashboard") is not True:
            with open(f"{cachedir}/prefect.dashboard","w") as f:
                f.write(f"{SERVER_DASHBOARD}")
        return 0
    else:
        print(f"⚠️ Server did not respond in time. Check logs at {LOG_FILE}")
        return 1


def stop_prefect_server():
    """
    Stop prefect server running in the current installation
    Note: it will only stop prefect server which is running from the current installation
    For this pipeline, default port (4250) is kept seperate from default prefect port 4200.
    """
    if not os.path.exists(pid_file):
        print("⚠️ No PID file found. Cannot stop Prefect server.")
        return
    try:
        with open(pid_file, "r") as f:
            pid = int(f.read().strip())
        print(f"🛑 Stopping Prefect server with PID {pid} ...")
        os.kill(pid, signal.SIGTERM)
        os.remove(pid_file)
        print("✅ Server stopped and PID file removed.")
    except ProcessLookupError:
        print(f"⚠️ No such process with PID {pid}. Removing stale PID file.")
        os.remove(pid_file)
    except Exception as e:
        print(f"❌ Error stopping server: {e}")


def show_prefect_config():
    """
    Print the effective Prefect config in this environment.
    """
    print("🔍 Prefect config in current environment:")
    subprocess.run(["prefect", "config", "view"], env=get_prefect_env())

# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if isinstance(obj, types.FunctionType) and obj.__module__ == __name__
]



import os
import asyncio
import threading
import types
from pathlib import Path
from prefect.client.orchestration import get_client
from prefect.client.schemas.sorting import LogSort
from prefect.client.schemas.filters import LogFilter


async def save_logs_by_task_id(task_run_id, task_name, logfile, poll_interval=5, stop_event=None):
    """
    Fetch and save prefect task logs to a file
    
    Parameters
    ----------
    task_run_id : str
        The Prefect task run ID to monitor
    task_name : str
        Task name
    logfile : str
        Output log file
    poll_interval : int
        How often to check for new logs (in seconds)
    stop_event : threading.Event
        Optional external signal to stop logging
    """
    seen_ids = set()
    while not (stop_event and stop_event.is_set()):
        try:
            async with get_client() as client:
                log_filter = LogFilter(task_run={"any_": [task_run_id]})
                logs = await client.read_logs(log_filter=log_filter,sort=LogSort.TIMESTAMP_ASC)
                with open(logfile, "a") as f:
                    for log in logs:
                        if log.id not in seen_ids:
                            seen_ids.add(log.id)
                            ts = log.timestamp.to_datetime_string()
                            if str(log.task_run_id)==str(task_run_id):
                                f.write(f"{ts} | {task_name} | {log.message}\n")
        except Exception as e:
            with open(logfile, "a") as f:
                f.write(f"Error fetching logs: {e}\n")
        await asyncio.sleep(poll_interval)
 
        
async def save_logs_by_flow_id(flow_run_id, flow_name, logfile, poll_interval=5, stop_event=None):
    """
    Fetch and save prefect flow logs to a file
    
    Parameters
    ----------
    flow_run_id : str
        The Prefect flow run ID to monitor
    flow_name : str
        Flow name
    logfile : str
        Output log file
    poll_interval : int
        How often to check for new logs (in seconds)
    stop_event : threading.Event
        Optional external signal to stop logging
    """
    seen_ids = set()
    while not (stop_event and stop_event.is_set()):
        try:
            async with get_client() as client:
                log_filter = LogFilter(flow_run={"any_": [flow_run_id]})
                logs = await client.read_logs(log_filter=log_filter, sort=LogSort.TIMESTAMP_ASC)

                with open(logfile, "a") as f:
                    for log in logs:
                        if log.id not in seen_ids:
                            seen_ids.add(log.id)
                            # Only include logs without task_run_id = flow-level logs
                            if log.task_run_id is None:
                                ts = log.timestamp.to_datetime_string()
                                f.write(f"{ts} | {flow_name} | {log.message}\n")

        except Exception as e:
            with open(logfile, "a") as f:
                f.write(f"Error fetching flow logs: {e}\n")

        await asyncio.sleep(poll_interval)

def start_log_task_saver(task_run_id, task_name, logfile, poll_interval=5, stop_event=None):
    """
    Start a background thread that saves Prefect task logs to a file continuously.

    Parameters
    ----------
    task_run_id : str
        The Prefect task run ID to monitor
    task_name : str
        Task name
    logfile : str
        Output log file.
    poll_interval : int
        How often to check for new logs (in seconds)
    stop_event : threading.Event
        Optional external signal to stop logging
    """
    def run_loop():
        asyncio.run(save_logs_by_task_id(task_run_id, task_name, logfile, poll_interval, stop_event))

    thread = threading.Thread(target=run_loop, daemon=True)
    thread.start()
    return thread


def start_flow_log_saver(flow_run_id, flow_name, logfile, poll_interval=5, stop_event=None):
    """
    Start a background thread that saves Prefect flow logs to a file continuously.
    
    Parameters
    ----------
    flow_run_id : str
        The Prefect flow run ID to monitor
    flow_name : str
        Flow name
    logfile : str
        Output log file
    poll_interval : int
        How often to check for new logs (in seconds)
    stop_event : threading.Event
        Optional external signal to stop logging
    """
    def run_loop():
        asyncio.run(save_logs_by_flow_id(flow_run_id, flow_name, logfile, poll_interval, stop_event))

    thread = threading.Thread(target=run_loop, daemon=True)
    thread.start()
    return thread

  
# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if isinstance(obj, types.FunctionType) and obj.__module__ == __name__
]




import logging
import psutil
import dask
import numpy as np
import argparse
import traceback
import time
import sys
import os
from casatasks import casalog
from casatools import msmetadata
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)


try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def chanlist_to_str(lst):
    lst = sorted(lst)
    ranges = []
    start = lst[0]
    for i in range(1, len(lst)):
        if lst[i] != lst[i - 1] + 1:
            if lst[i - 1] > start:
                ranges.append(f"{start}~{lst[i - 1]}")
            elif lst[i - 1] == start:
                ranges.append(f"{start}")
            start = lst[i]
    if lst[-1] > start:
        ranges.append(f"{start}~{lst[-1]}")
    elif lst[-1] == start:
        ranges.append(f"{start}")
    return ";".join(ranges)


def split_target_scans(
    msname,
    workdir,
    timeres,
    freqres,
    datacolumn,
    spw="",
    spectral_chunk=-1,
    n_spectral_chunk=-1,
    scans=[],
    prefix="targets",
    time_interval=-1,
    time_window=-1,
    quack_timestamps=-1,
    merge_spws=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    max_cpu_frac=0.8,
    max_mem_frac=0.8,
    dask_addr=None,
):
    """
    Split target scans

    Parameters
    ----------
    msname : str
        Measurement set
    workdir : str
        Work directory
    timeres : float
        Time resolution in seconds
    freqres : float
        Frequency resolution in MHz
    datacolumn : str
        Data column to split
    spw : str, optional
        Spectral window
    spectral_chunk : float, optional
        Spectral chunk in MHz
    n_spectral_chunk : int, optional
        Number of spectral chunks to split from the beginning
    scans : list
        Scan list to split
    prefix : str, optional
        Splited ms prefix
    time_interval : float
        Time interval in seconds
    time_window : float
        Time window in seconds
    quack_timestamps : int, optional
        Number of timestamps ignored at the start and end of each scan
    merge_spws : bool, optional
        Merge spectral window ranges
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    max_cpu_frac : float, optional
        Maximum CPU fraction to use
    max_mem_frac : float, optional
        Maximum memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    list
        Splited ms list
    """
    start_time = time.time()
    try:
        os.chdir(workdir)
        print(f"Spliting ms : {msname}")
        target_scans, cal_scans, f_scans, g_scans, p_scans = get_cal_target_scans(
            msname
        )
        valid_scans = get_valid_scans(msname)
        filtered_scan_list = []
        for scan in target_scans:
            if scan in valid_scans:
                if len(scans) == 0 or (len(scans) > 0 and scan in scans):
                    filtered_scan_list.append(scan)
        filtered_scan_list = sorted(filtered_scan_list)

        #######################################
        # Extracting time frequency information
        #######################################
        msmd = msmetadata()
        msmd.open(msname)
        chanres = msmd.chanres(0, unit="MHz")[0]
        freqs = msmd.chanfreqs(0, unit="MHz")
        bw = max(freqs) - min(freqs)
        nchan = msmd.nchan(0)
        msmd.close()
        if freqres > 0:  # Image resolution is in MHz
            chanwidth = int(freqres / chanres)
            if chanwidth < 1:
                chanwidth = 1
        else:
            chanwidth = 1
        if timeres > 0:  # Image resolution is in seconds
            timebin = str(timeres) + "s"
        else:
            timebin = ""

        #############################
        # Making spectral chunks
        #############################
        bad_spws = get_bad_chans(msname).split("0:")[-1].split(";")
        # Derive good spectral windows by finding gaps between bad SPWs
        good_spws_list = []
        for i in range(len(bad_spws) - 1):
            start_chan = int(bad_spws[i].split("~")[-1]) + 1
            end_chan = int(bad_spws[i + 1].split("~")[0]) - 1
            good_spws_list.append(f"{start_chan}~{end_chan}")

        # If no gaps found, fall back to full good channels
        if good_spws_list:
            good_spws = "0:" + ";".join(good_spws_list)
        else:
            good_spws = get_good_chans(msname)

        # Intersect with user-specified SPW if provided
        spw_suffix = good_spws.split("0:")[-1].split(";")
        if spw:
            common_spws = get_common_spw(good_spws, spw)
            if common_spws != "":
                good_spws = common_spws.split("0:")[-1].split(";")
            else:
                good_spws = spw_suffix
        else:
            good_spws = spw_suffix
        if len(good_spws) == 0:
            print("No good spectral window is found.")
            return 1, []

        #############################
        chanlist = []
        if spectral_chunk > 0:
            if spectral_chunk > bw:
                print(
                    f"Given spectral chunk: {spectral_chunk} is more than total bandwidth: {bw} MHz."
                )
                spectral_chunk = bw
            nchan_per_chunk = max(1, int(spectral_chunk / chanres))
            good_channels = []
            for good_spw in good_spws:
                start_chan = int(good_spw.split("~")[0])
                end_chan = int(good_spw.split("~")[-1])
                for s in range(start_chan, end_chan):
                    good_channels.append(s)
            channel_chunks = split_into_chunks(good_channels, nchan_per_chunk)
            for chunk in channel_chunks:
                chan_str = chanlist_to_str(chunk)
                if chan_str not in chanlist:
                    chanlist.append(chan_str)
            if n_spectral_chunk > 0:
                indices = np.linspace(
                    0, len(chanlist) - 1, num=n_spectral_chunk, dtype=int
                )
                chanlist = [chanlist[i] for i in indices]
        else:
            chan_range = ""
            for good_spw in good_spws:
                s = int(good_spw.split("~")[0])
                e = int(good_spw.split("~")[-1])
                chan_range += f"{s}~{e};"
            chan_range = chan_range[:-1]
            if chan_range not in chanlist:
                chanlist.append(chan_range)

        if merge_spws:
            temp_spw = ";".join(chanlist)
            chanlist = [temp_spw]

        print(f"Spliting channel blocks : {chanlist}")

        ##################################
        # Parallel spliting
        ##################################
        if len(chanlist) > 0:
            total_chunks = len(chanlist) * len(filtered_scan_list)
        else:
            total_chunks = len(filtered_scan_list)

        #############################################
        # Memory limit
        #############################################
        mem_limit = single_mstransform(dry_run=True)
        #######################
        _, _, max_n_jobs, n_threads, mem_limit, dask_dir = (
            get_dask_client(
                total_chunks,
                dask_dir=workdir,
                cpu_frac=max_cpu_frac,
                mem_frac=max_mem_frac,
                min_mem_per_job=mem_limit,
                only_cal=True,
            )
        )
        os.system(f"rm -rf {dask_dir}")
        tasks = []
        splited_ms_list = []
        for scan in filtered_scan_list:
            timerange_list = get_timeranges_for_scan(
                msname,
                scan,
                time_interval,
                time_window,
                quack_timestamps=quack_timestamps,
            )
            timerange = ",".join(timerange_list)
            for chanrange in chanlist:
                chanrange_str = (
                    chanrange.split(";")[0].split("~")[0]
                    + "~"
                    + chanrange.split(";")[-1].split("~")[-1]
                )
                outputvis = f"{workdir}/{prefix}_scan_{scan}_spw_{chanrange_str}.ms"
                if os.path.exists(f"{outputvis}/.splited"):
                    print(f"{outputvis} is already splited successfully.")
                    splited_ms_list.append(outputvis)
                else:
                    task = delayed(single_mstransform)(
                        msname=msname,
                        outputms=outputvis,
                        field="",
                        scan=scan,
                        width=chanwidth,
                        timebin=timebin,
                        datacolumn="DATA",
                        spw="0:" + chanrange,
                        corr="",
                        timerange=timerange,
                        n_threads=n_threads,
                        dry_run=False,
                    )
                    tasks.append(task)
        #####################################
        # Adaptive dask client
        #####################################
        if cpu_frac == max_cpu_frac and mem_frac == max_mem_frac:
            total_chunks = len(tasks)
            if total_chunks > 0:
                if dask_addr is None:
                    dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                        get_dask_client(
                            total_chunks,
                            dask_dir=workdir,
                            cpu_frac=cpu_frac,
                            mem_frac=mem_frac,
                            min_mem_per_job=mem_limit,
                        )
                    )
                else:
                    _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                        get_dask_client(
                            total_chunks,
                            dask_dir=workdir,
                            cpu_frac=cpu_frac,
                            mem_frac=mem_frac,
                            min_mem_per_job=mem_limit,
                            only_cal=True,
                        )
                    )
                    dask_client=Client(address=dask_addr)
                    os.system(f"rm -rf {dask_dir}")
                wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
                futures = dask_client.compute(tasks)
                results = list(dask_client.gather(futures))
                dask_client.close()
                if dask_addr is None:
                    dask_cluster.close()
                    os.system(f"rm -rf {dask_dir}")
                for r in results:
                    splited_ms_list.append(r)
        else:
            while True:
                total_chunks = len(tasks)
                if total_chunks == 0:
                    break
                else:
                    if dask_addr is None:
                        (
                            dask_client,
                            dask_cluster,
                            n_jobs,
                            n_threads,
                            mem_limit,
                            dask_dir,
                        ) = get_dask_client(
                            total_chunks,
                            dask_dir=workdir,
                            cpu_frac=cpu_frac,
                            mem_frac=mem_frac,
                            min_mem_per_job=mem_limit,
                        )
                    else:
                        (
                            _,
                            _,
                            n_jobs,
                            n_threads,
                            mem_limit,
                            dask_dir,
                        ) = get_dask_client(
                            total_chunks,
                            dask_dir=workdir,
                            cpu_frac=cpu_frac,
                            mem_frac=mem_frac,
                            min_mem_per_job=mem_limit,
                            only_cal=True,
                        )
                        dask_client=Client(address=dask_addr)
                        os.system(f"rm -rf {dask_dir}")
                    wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
                    chunk_tasks = tasks[0 : min(n_jobs, max_n_jobs)]
                    for ctask in chunk_tasks:
                        tasks.remove(ctask)
                    futures = dask_client.compute(tasks)
                    results = list(dask_client.gather(futures))
                    dask_client.close()
                    if dask_addr is None:
                        dask_cluster.close()
                        os.system(f"rm -rf {dask_dir}")
                    for r in results:
                        splited_ms_list.append(r)
                    n_current_process = (
                        get_nprocess_solarpipe(workdir) - 1
                    )  # One is subtracted for the current process
                    if len(tasks) == 0:
                        break
                    elif n_current_process == 0:
                        available_cpu_frac = round(
                            (100 - psutil.cpu_percent(interval=1)) / 100.0, 2
                        )
                        available_mem_frac = round(
                            psutil.virtual_memory().available
                            / psutil.virtual_memory().total,
                            2,
                        )
                        cpu_frac = min(max_cpu_frac, max(cpu_frac, available_cpu_frac))
                        mem_frac = min(max_mem_frac, max(mem_frac, available_mem_frac))
                        print(
                            f"Updated CPU fraction: {cpu_frac}, memory fraction: {mem_frac}."
                        )
        print("##################")
        print("Spliting of target scans are done successfully.")
        print("Total time taken : ", time.time() - start_time)
        print("##################\n")
        return 0, splited_ms_list
    except Exception as e:
        traceback.print_exc()
        print("##################")
        print("Spliting of target scans are unsuccessful.")
        print("Total time taken : ", time.time() - start_time)
        print("##################\n")
        return 1, []


def main(
    msname,
    workdir="",
    datacolumn="data",
    spw="",
    scans="",
    time_window=-1,
    time_interval=-1,
    quack_timestamps=-1,
    spectral_chunk=-1,
    n_spectral_chunk=-1,
    freqres=-1,
    timeres=-1,
    prefix="targets",
    merge_spws=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    max_cpu_frac=0.8,
    max_mem_frac=0.8,
    logfile=None,
    jobid=0,
    start_remote_log=False,
    dask_addr=None,
):
    """
    Split target scans from a measurement set into smaller chunks for parallel processing.

    Parameters
    ----------
    msname : str
        Path to the input measurement set (MS).
    workdir : str, optional
        Working directory for intermediate and output products. If empty, defaults to `<msname>/workdir`.
    datacolumn : str, optional
        Column of the MS to use for splitting (e.g., "DATA", "CORRECTED"). Default is "data".
    spw : str, optional
        Spectral windows to include, in CASA syntax (e.g., "0~3"). Default is "" (all).
    scans : str, optional
        Scan numbers to include, comma-separated (e.g., "1,3,5"). Default is "" (all).
    time_window : float, optional
        Time window in seconds to group scans. Set -1 to disable. Default is -1.
    time_interval : float, optional
        Integration time interval in seconds for time averaging. Set -1 to disable. Default is -1.
    quack_timestamps : float, optional
        Time in seconds to flag at the beginning of each scan ("quack"). -1 to disable. Default is -1.
    spectral_chunk : float, optional
        Width of spectral chunks in MHz. If set > 0, overrides `n_spectral_chunk`. Default is -1.
    n_spectral_chunk : int, optional
        Number of spectral chunks to split each SPW into. Ignored if `spectral_chunk` > 0. Default is -1.
    freqres : float, optional
        Frequency resolution in MHz for spectral averaging. Set -1 to disable. Default is -1.
    timeres : float, optional
        Time resolution in seconds for time averaging. Set -1 to disable. Default is -1.
    prefix : str, optional
        Prefix for the output split MS files. Default is "targets".
    merge_spws : bool, optional
        If True, merge all SPWs before splitting. Default is False.
    cpu_frac : float, optional
        Fraction of available CPUs to allocate per task. Default is 0.8.
    mem_frac : float, optional
        Fraction of available memory to allocate per task. Default is 0.8.
    max_cpu_frac : float, optional
        Maximum total CPU usage across all tasks. Default is 0.8.
    max_mem_frac : float, optional
        Maximum total memory usage across all tasks. Default is 0.8.
    logfile : str or None, optional
        Path to log file. If None, logging to file is disabled. Default is None.
    jobid : int, optional
        Job identifier for tracking and PID storage. Default is 0.
    start_remote_log : bool, optional
        If True, enables remote logging using credentials stored in workdir. Default is False.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "do_target_split", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ###########

    try:
        if msname and os.path.exists(msname):
            print("\n###################################")
            print("Start spliting target scans.")
            print("###################################\n")
            scans = [int(i) for i in scans.split(",")] if scans else []
            msg, final_target_mslist = split_target_scans(
                msname,
                workdir,
                float(timeres),
                float(freqres),
                datacolumn,
                spw=str(spw),
                time_window=float(time_window),
                time_interval=float(time_interval),
                quack_timestamps=int(quack_timestamps),
                scans=scans,
                n_spectral_chunk=int(n_spectral_chunk),
                prefix=prefix,
                merge_spws=merge_spws,
                spectral_chunk=float(spectral_chunk),
                cpu_frac=float(cpu_frac),
                mem_frac=float(mem_frac),
                max_cpu_frac=float(max_cpu_frac),
                max_mem_frac=float(max_mem_frac),
                dask_addr=dask_addr,
            )
        else:
            print("Please provide correct measurement set.\n")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Split target scans", formatter_class=SmartDefaultsHelpFormatter
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "msname",
        type=str,
        help="Name of measurement set (required positional argument)",
    )
    basic_args.add_argument(
        "--workdir",
        type=str,
        default="",
        help="Name of work directory",
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--datacolumn",
        type=str,
        default="data",
        help="Data column to split",
    )
    adv_args.add_argument(
        "--spw", type=str, default="", help="Spectral window to split"
    )
    adv_args.add_argument(
        "--scans",
        type=str,
        default="",
        help="Target scan list (default: all)",
        metavar="SCANS (Comma seperated)",
    )
    adv_args.add_argument(
        "--time_window",
        type=float,
        default=-1,
        help="Time window in seconds",
        metavar="Float",
    )
    adv_args.add_argument(
        "--time_interval",
        type=float,
        default=-1,
        help="Time interval in seconds",
        metavar="Float",
    )
    adv_args.add_argument(
        "--quack_timestamps",
        type=int,
        default=-1,
        help="Time stamps to ignore at the start and end of the each scan",
        metavar="Integer",
    )
    adv_args.add_argument(
        "--spectral_chunk",
        type=float,
        default=-1,
        help="Spectral chunk in MHz",
        metavar="Float",
    )
    adv_args.add_argument(
        "--n_spectral_chunk",
        type=int,
        default=-1,
        help="Numbers of spectral chunks to split",
        metavar="Integer",
    )
    adv_args.add_argument(
        "--freqres",
        type=float,
        default=-1,
        help="Frequency to average in MHz",
        metavar="Float",
    )
    adv_args.add_argument(
        "--timeres",
        type=float,
        default=-1,
        help="Time bin to average in seconds",
        metavar="Float",
    )
    adv_args.add_argument(
        "--prefix",
        type=str,
        default="targets",
        help="Splited ms prefix name",
    )
    adv_args.add_argument(
        "--merge_spws", action="store_true", help="Merge spectral windows"
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="CPU fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Memory fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--max_cpu_frac",
        type=float,
        default=0.8,
        help="Maximum CPU fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--max_mem_frac",
        type=float,
        default=0.8,
        help="Maximum memory fraction to use",
        metavar="Float",
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Log file")
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        msname=args.msname,
        workdir=args.workdir,
        datacolumn=args.datacolumn,
        spw=args.spw,
        scans=args.scans,
        time_window=args.time_window,
        time_interval=args.time_interval,
        quack_timestamps=args.quack_timestamps,
        spectral_chunk=args.spectral_chunk,
        n_spectral_chunk=args.n_spectral_chunk,
        freqres=args.freqres,
        timeres=args.timeres,
        prefix=args.prefix,
        merge_spws=args.merge_spws,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        max_cpu_frac=args.max_cpu_frac,
        max_mem_frac=args.max_mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nSpliting target scans are done.\n###################\n"
    )
    os._exit(result)
